import hashlib
import json
import logging
import os
import threading
from functools import partial
from collections import namedtuple

import requests
from more_executors import Executors
from more_executors.futures import f_map, f_flat_map, f_return, f_proxy, f_sequence
from io import StringIO
from ..compat_attr import evolve
from ..model.repository.repo_lock import LOCK_CLAIM_STR
from ..page import Page
from ..criteria import Criteria
from ..model import (
    Repository,
    FileRepository,
    YumRepository,
    MaintenanceReport,
    Distributor,
    Unit,
    Task,
)
from ..log import TimedLogger
from ..util import dict_put, naturalsize
from .search import search_for_criteria
from .errors import PulpException
from .poller import TaskPoller
from . import retry

from .ud_mappings import compile_ud_mappings
from .copy import CopyOptions


LOG = logging.getLogger("pubtools.pulplib")

UploadResult = namedtuple("UploadResult", ["checksum", "size"])


class Client(object):
    """A client for the Pulp 2.x API.

    This class provides high-level methods for querying a Pulp server and performing
    actions in Pulp.

    **Usage of futures:**

    Most values are returned as :class:`~concurrent.futures.Future` objects.
    Returned futures are resolved when:

    - a request to Pulp succeeded, if operation is synchronous (e.g. get, search)
    - or a request to Pulp succeeded and any spawned tasks succeeded, if operation
      is asynchronous (e.g. publish)

    If a future is currently awaiting one or more Pulp tasks, cancelling the future
    will attempt to cancel those tasks.

    **Client lifecycle:**

    .. versionadded:: 2.12.0

    Client instances support the context manager protocol and can be used
    via a ``with`` statement, as in example:

    .. code-block:: python

        with Client(url="https://pulp.example.com/") as client:
            do_something_with(client)

    While not mandatory, it is encouraged to ensure that any threads associated with
    the client are promptly shut down.

    **Proxy futures:**

    .. versionadded:: 2.1.0

    All :class:`~concurrent.futures.Future` objects produced by this client are
    *proxy futures*, meaning that attribute and method lookups on the objects are
    proxied to the future's result, blocking as needed.

    This allows the client to be used within blocking code without having to
    scatter calls to ``.result()`` throughout.

    For example, this block of code:

    .. code-block:: python

        repo = client.get_repository(repo_id).result()
        publish_tasks = repo.publish().result()
        task_ids = ','.join([t.id for t in publish_tasks])
        log.info("Published %s: %s", repo.id, task_ids)

    ...may be alternatively written without the calls to ``.result()``, due to
    the usage of proxy futures:

    .. code-block:: python

        repo = client.get_repository(repo_id)
        publish_tasks = repo.publish()
        task_ids = ','.join([t.id for t in publish_tasks])
        log.info("Published %s: %s", repo.id, task_ids)

    **Retries:**

    In general, for all methods which represent an idempotent operation and
    which return a Future, the library may retry operations several times in case
    of failure.

    **Throttling:**

    This client will internally throttle the number of Pulp tasks allowed to be
    outstanding at any given time, so that no single client can overload the Pulp
    task queue.
    """

    # Various defaults can be controlled by environment variables.
    # These are not a documented/supported feature of the library.
    # They're more of an emergency escape hatch to temporarily resolve issues.
    _REQUEST_THREADS = int(os.environ.get("PUBTOOLS_PULPLIB_REQUEST_THREADS", "4"))
    _UPLOAD_THREADS = int(os.environ.get("PUBTOOLS_PULPLIB_UPLOAD_THREADS", "4"))
    _PAGE_SIZE = int(os.environ.get("PUBTOOLS_PULPLIB_PAGE_SIZE", "2000"))
    _TASK_THROTTLE = int(os.environ.get("PUBTOOLS_PULPLIB_TASK_THROTTLE", "200"))
    _CHUNK_SIZE = int(os.environ.get("PUBTOOLS_PULPLIB_CHUNK_SIZE", 1024 * 1024 * 10))

    # Policy used when deciding whether to retry operations.
    # This is mainly provided here as a hook for autotests, so the policy can be
    # overridden there.
    _RETRY_POLICY = retry.PulpRetryPolicy()

    def __init__(self, url, **kwargs):
        """
        Args:
            str url
                URL of a Pulp server, e.g. https://pulp.example.com/.

            int task_throttle
                Maximum number of queued or running tasks permitted for this client.
                If more than this number of tasks are running, the client will wait before triggering more.
                This can be used to ensure no single client overwhelms the Pulp server.

            int threads
                A hint for the number of threads the client should use in order to
                make concurrent requests to Pulp.

                This number should be considered as more of a scaling factor than
                a precise thread count. The client in practice will use more than
                this number of threads.

            object auth, cert, headers, max_redirects, params, proxies, verify
                Any of these arguments, if provided, are used to initialize
                :class:`requests.Session` objects used by the client.

                These may be used, for example, to configure the credentials
                for the Pulp server or to use an alternative CA bundle.

        .. versionadded:: 2.31.0
            Added the ``threads`` argument.
        """
        self._url = url

        while self._url.endswith("/"):
            self._url = self._url[:-1]

        self._session_kwargs = {}
        for arg in (
            "auth",
            "cert",
            "headers",
            "max_redirects",
            "params",
            "proxies",
            "verify",
        ):
            if arg in kwargs:
                self._session_kwargs[arg] = kwargs.pop(arg)

        task_throttle = kwargs.pop("task_throttle", self._TASK_THROTTLE)
        threads = kwargs.pop("threads", None)
        if threads is not None and threads < 1:
            threads = 1

        if kwargs:
            raise TypeError(
                "Unexpected keyword argument(s) %s" % ",".join(kwargs.keys())
            )

        self._tls = threading.local()

        # Cache supported types, used during searches.
        self._server_type_ids = None

        # executor only for issuing HTTP requests to Pulp:
        # - does not cover watching Pulp tasks
        # - checks HTTP response and raises if it's not JSON or
        #   it's not successful
        self._request_executor = (
            Executors.thread_pool(
                name="pubtools-pulplib-requests",
                max_workers=threads or self._REQUEST_THREADS,
            )
            .with_map(self._unpack_response)
            .with_retry(retry_policy=self._RETRY_POLICY)
        )

        # executor for uploads:
        # - reading of file content for upload, and calculation of checksums,
        #   happens in this executor
        # - HTTP requests don't happen here and are still submitted via
        #   request_executor (hence no retry needed on this executor)
        self._upload_executor = Executors.thread_pool(
            name="pubtools-pulplib-uploads", max_workers=threads or self._UPLOAD_THREADS
        )

        # executor for issuing HTTP requests to Pulp which spawn tasks:
        # - checks HTTP response and raises if it's not JSON or
        #   it's not successful
        # - unpacks spawned_tasks from response to find task IDs to watch
        # - waits for tasks to complete
        # - retries whole thing (resubmitting request & making a new task) if needed
        # - throttles number of tasks pending
        poller = TaskPoller(self._new_session(), self._url)
        self._task_executor = (
            Executors.thread_pool(
                name="pubtools-pulplib-tasks",
                max_workers=threads or self._REQUEST_THREADS,
            )
            .with_map(self._unpack_response)
            .with_map(self._log_spawned_tasks)
            .with_poll(poller, cancel_fn=poller.cancel)
            .with_throttle(task_throttle)
            .with_retry(retry_policy=self._RETRY_POLICY)
        )

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._request_executor.__exit__(*args, **kwargs)
        self._upload_executor.__exit__(*args, **kwargs)
        self._task_executor.__exit__(*args, **kwargs)

    def get_repository(self, repository_id):
        """Get a repository by ID.

        Args:
            repository_id (str)
                The ID of the repository to be fetched.

        Returns:
            Future[:class:`~pubtools.pulplib.Repository`]
                A future holding the fetched Repository.
        """
        page_f = self.search_repository(Criteria.with_id(repository_id))

        def unpack_page(page):
            if len(page.data) != 1:
                raise PulpException("Repository id=%s was not found" % repository_id)
            return page.data[0]

        repo_f = f_map(page_f, unpack_page)
        return f_proxy(repo_f)

    def search_repository(self, criteria=None):
        """Search for repositories matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all repositories.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Repository` objects.
        """
        search_options = {"distributors": True, "importers": True}
        return self._search(
            Repository, "repositories", criteria=criteria, search_options=search_options
        )

    def search_content(self, criteria=None):
        """Search for units across all repositories.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all units.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Unit` subclasses objects.

        .. versionadded:: 2.6.0
        """
        # Criteria will be serialized into a Pulp search at the time we
        # actually do the query, but validate eagerly as well so we raise
        # ASAP on invalid input.
        search_for_criteria(criteria, Unit, None)

        if self._server_type_ids is None:
            # We'll be using this in a moment either to set default IDs or to
            # reject searches for invalid types.
            # Note: no locking, so if we're called from multiple threads we in
            # theory might waste some time querying the types more than once.
            self._server_type_ids = self.get_content_type_ids()

        return f_proxy(
            f_flat_map(
                self._server_type_ids,
                lambda ids: self._search_content_with_server_type_ids(criteria, ids),
            )
        )

    def copy_content(
        self, from_repository, to_repository, criteria=None, options=CopyOptions()
    ):
        """Copy content from one repository to another.

        Args:
            from_repository (:class:`~pubtools.pulplib.Repository`)
                A repository used as the source of content.

            to_repository (:class:`~pubtools.pulplib.Repository`)
                A repository to receive copied content.

            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used to find units for copy.
                If None, all units in the source repo will be copied.

            options (:class:`~pubtools.pulplib.CopyOptions`)
                Options influencing the copy.

                Some options may be specific to certain repository and content
                types. Options which are not applicable to this copy will be
                ignored.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when the copy completes.

                The :meth:`~pubtools.pulplib.Task.units` attribute may be inspected
                to determine which unit(s) were copied. Note that the returned units
                typically will have only a subset of available fields.

        .. versionadded:: 2.17.0

        .. versionadded:: 2.29.0
            Added the ``options`` argument.
        """

        raw_options = {}

        if (
            isinstance(to_repository, YumRepository)
            and options.require_signed_rpms is not None
        ):
            dict_put(
                raw_options,
                "override_config.require_signature",
                options.require_signed_rpms,
            )

        return f_proxy(
            self._do_associate(
                from_repository.id, to_repository.id, criteria, raw_options
            )
        )

    def update_content(self, unit):
        """Update mutable fields on an existing unit.

        Args:
            unit (:class:`~pubtools.pulplib.Unit`)
                A unit to be updated.

                This unit must have a known ``unit_id``.

                Only those fields documented as *mutable* will have any effect during
                the update (e.g. ``cdn_path``). Other fields either cannot be updated
                at all, or can only be updated by re-uploading the associated content.

        Returns:
            Future
                A future which is resolved with a value of ``None`` once the unit
                has been updated.

        .. versionadded:: 2.20.0
        """

        if not unit.unit_id:
            raise ValueError("unit_id missing on call to update_content()")

        url = os.path.join(
            self._url,
            "pulp/api/v2/content/units/%s/%s/pulp_user_metadata/"
            % (unit.content_type_id, unit.unit_id),
        )

        out = self._request_executor.submit(
            self._do_request, method="PUT", url=url, json=unit._usermeta
        )

        # The Pulp API is defined as returning 'null' so this should be a no-op,
        # but to be extra sure we don't return anything unexpected, we'll force
        # the return value to None.
        out = f_map(out, lambda _: None)

        return out

    def update_repository(self, repository):
        """Update mutable fields on an existing repository.

        Args:
            repository (:class:`~pubtools.pulplib.Repository`)
                A repository to be updated.

                Only those fields documented as *mutable* will have any effect during
                the update (e.g. ``product_versions``). Other fields cannot be updated
                using this library.

        Returns:
            Future
                A future which is resolved with a value of ``None`` once the
                repository has been updated.

        .. versionadded:: 2.29.0
        """

        url = os.path.join(self._url, "pulp/api/v2/repositories/%s/" % repository.id)

        update = {"delta": {"notes": repository._mutable_notes}}
        out = self._request_executor.submit(
            self._do_request, method="PUT", url=url, json=update
        )

        # The Pulp API may actually return an updated version of the repository,
        # but for consistency with update_content we won't return it.
        # The caller can re-fetch if desired.
        out = f_map(out, lambda _: None)

        return out

    def _search_content_with_server_type_ids(self, criteria, server_type_ids):
        prepared_search = search_for_criteria(criteria, Unit, None)
        type_ids = prepared_search.type_ids
        if not type_ids:
            type_ids = server_type_ids
        missing = set(type_ids) - set(server_type_ids)
        if missing:
            raise ValueError(
                "Content type: %s is not supported by server" % ",".join(missing)
            )

        return self._search(
            Unit,
            ["content/units/%s" % x for x in type_ids],
            criteria=criteria,
            search_options={"include_repos": True},
        )

    def search_distributor(self, criteria=None):
        """Search the distributors matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all distributors.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Distributor` objects.

        .. versionadded:: 2.1.0
        """
        return self._search(Distributor, "distributors", criteria=criteria)

    def search_task(self, criteria=None):
        """Search the tasks matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all tasks.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Task` objects.

        .. versionadded:: 2.19.0
        """
        return self._search(Task, "tasks", criteria=criteria)

    def _search(
        self,
        return_type,
        resource_types,
        search_type="search",
        search_options=None,
        criteria=None,
    ):  # pylint:disable = too-many-arguments

        if not isinstance(resource_types, (list, tuple)):
            resource_types = [resource_types]

        responses = []
        searches = []
        urls = []
        for resource_type in resource_types:
            url = os.path.join(
                self._url, "pulp/api/v2/%s/%s/" % (resource_type, search_type)
            )
            urls.append(url)
            prepared_search = search_for_criteria(criteria, return_type)

            search = {
                "criteria": {
                    "skip": 0,
                    "limit": self._PAGE_SIZE,
                    "filters": prepared_search.filters,
                }
            }

            if prepared_search.unit_fields is not None:
                search["criteria"]["fields"] = sorted(
                    set(
                        [field.pulp_field_name for field in prepared_search.unit_fields]
                    )
                )

            search.update(search_options or {})

            if search_type == "search/units":
                # Unit searches need a little special handling:
                # - serialization might have extracted some type_ids
                # - filters/fields should be wrapped under 'unit'
                #   (we do not support searching on associations right now)
                if prepared_search.type_ids:
                    search["criteria"]["type_ids"] = prepared_search.type_ids

                for elem in ("filters", "fields"):
                    if elem in search["criteria"]:
                        search["criteria"][elem] = {"unit": search["criteria"][elem]}

            searches.append(search)
            responses.append(self._do_search(url, search))

        # When this request is resolved, we'll have the first page of data.
        # We'll need to convert that into a page and also keep going with
        # the search if there's more to be done.
        return f_proxy(
            f_map(
                f_sequence(responses),
                lambda data: self._handle_page(urls, return_type, searches, data),
            )
        )

    def _search_repo_units(self, repo_id, criteria):
        resource_type = "repositories/%s" % repo_id

        return self._search(
            Unit, resource_type, search_type="search/units", criteria=criteria
        )

    def get_maintenance_report(self):
        """Get the current maintenance mode status for this Pulp server.

        Returns:
            Future[:class:`~pubtools.pulplib.MaintenanceReport`]
                A future describes the maintenance status

        .. versionadded:: 1.4.0
        """
        report_ft = self._do_get_maintenance()

        return f_map(
            report_ft,
            lambda data: MaintenanceReport._from_data(data)
            if data
            else MaintenanceReport(),
        )

    def set_maintenance(self, report):
        """Set maintenance mode for this Pulp server.

        Args:
            report:
                An updated :class:`~pubtools.pulplib.MaintenanceReport` object that
                will be used as the newest maintenance report.
        Return:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when maintenance mode has been updated successfully.

                The future contains a task triggered and awaited during the publish
                maintenance repository operation.

        .. versionadded:: 1.4.0
        """
        report_json = json.dumps(report._export_dict(), indent=4, sort_keys=True)
        report_fileobj = StringIO(report_json)

        repo = self.get_repository("redhat-maintenance").result()

        # upload updated report to repository and publish
        upload_ft = repo.upload_file(report_fileobj, "repos.json")

        return f_flat_map(upload_ft, lambda _: repo.publish())

    def get_content_type_ids(self):
        """Get the content types supported by this Pulp server.

        Returns:
            Future[list[str]]
                A future holding the IDs of supported content types.

                The returned values will depend on the plugins installed
                on the connected Pulp server.

                "modulemd", "rpm", "srpm" and "erratum" are some examples
                of common return values.

        .. versionadded:: 1.4.0
        """
        url = os.path.join(self._url, "pulp/api/v2/plugins/types/")

        out = self._request_executor.submit(self._do_request, method="GET", url=url)

        # The pulp API returns an object per supported type.
        # We only support returning the ID at this time.
        return f_proxy(f_map(out, lambda types: sorted([t["id"] for t in types])))

    def _do_upload_file(self, upload_id, file_obj, name="<unknown file>"):
        return self._upload_executor.submit(
            self._upload_file_loop, upload_id, file_obj, name
        )

    def _upload_file_loop(self, upload_id, file_obj, name):
        # Read a file in chunks, upload it to pulp under the given upload_id,
        # and return the checksum & bytes read.
        #
        # This method will block for the entire duration of the upload, which can
        # be slow for a large file. It is intended to be invoked only from within
        # the upload_executor in order to avoid blocking other operations.

        total_size = None

        is_file_object = "close" in dir(file_obj)
        if not is_file_object:
            # This is the preferred case (we're responsible for opening the file),
            # as we can then know the total expected size. (file-like objects in
            # general do not know their own size)
            total_size = os.path.getsize(file_obj)
            file_obj = open(file_obj, "rb")

        upload_logger = TimedLogger()
        checksum = hashlib.sha256()
        size = 0

        # We allow having a few chunks in flight at once, up to as many requests
        # as we can do in parallel.
        prev_chunks = [f_return() for _ in range(0, self._REQUEST_THREADS)]

        try:
            while True:
                data = file_obj.read(self._CHUNK_SIZE)

                if not data:
                    break

                if isinstance(data, str):
                    # if it's unicode, need to encode before calculate checksum
                    data = data.encode("utf-8")

                checksum.update(data)

                # Ensure the number of chunks are kept at a fixed size by waiting
                # for one of them to complete.
                prev_chunks.pop(0).result()

                # Then start upload of this chunk to Pulp.
                prev_chunks.append(self._do_upload(data, upload_id, size))

                size += len(data)

                # Log a message about the upload progress.
                if not total_size:
                    # no percentage can be calculated
                    pct = ""
                else:
                    pct = float(size) / total_size * 100
                    pct = " / %2d%%" % pct

                upload_logger.info(
                    "Still uploading %s: %s%s [%s]",
                    name,
                    naturalsize(size),
                    pct,
                    upload_id,
                )

            # No more data to read. As soon as all chunks in progress are done,
            # we have finished with the upload.
            for chunk in prev_chunks:
                chunk.result()

            return UploadResult(checksum.hexdigest(), size)

        finally:
            file_obj.close()

    def _publish_repository(self, repo, distributors_with_config):
        compiled = self._compile_notes(repo)

        tasks_f = f_map(compiled, lambda _: [])

        def do_next_publish(accumulated_tasks, distributor, config):
            distributor_tasks_f = self._publish_distributor(
                repo.id, distributor.id, config
            )
            return f_map(
                distributor_tasks_f,
                lambda distributor_tasks: accumulated_tasks + distributor_tasks,
            )

        for (distributor, config) in distributors_with_config:
            next_publish = partial(
                do_next_publish, distributor=distributor, config=config
            )
            tasks_f = f_flat_map(tasks_f, next_publish)

        return tasks_f

    def _publish_distributor(self, repo_id, distributor_id, override_config):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/publish/" % repo_id
        )
        body = {"id": distributor_id, "override_config": override_config}
        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _do_associate(self, src_repo_id, dest_repo_id, criteria=None, raw_options=None):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/associate/" % dest_repo_id
        )

        pulp_search = search_for_criteria(
            criteria, type_hint=Unit, unit_type_accum=None
        )

        body = {"source_repo_id": src_repo_id, "criteria": {}}
        if pulp_search.type_ids:
            body["criteria"]["type_ids"] = pulp_search.type_ids
        if pulp_search.filters:
            body["criteria"]["filters"] = {"unit": pulp_search.filters}

        body.update(raw_options or {})

        LOG.debug("Submitting %s associate: %s", url, body)

        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _do_unassociate(self, repo_id, criteria=None, limit=None):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/unassociate/" % repo_id
        )

        # use type hint=Unit so that if type_ids are the goal here
        # then we will get back a properly prepared PulpSearch with
        # a populated type_ids field
        pulp_search = search_for_criteria(
            criteria, type_hint=Unit, unit_type_accum=None
        )

        body = {"criteria": {"type_ids": pulp_search.type_ids}}

        if pulp_search.filters:
            # Filters to remove content are only effective when provided
            # with a type_id. Otherwise filters are skipped and all the
            # content is removed.
            if not pulp_search.type_ids:
                raise ValueError(
                    "Criteria to remove_content must specify at least one unit type!"
                )
            else:
                body["criteria"]["filters"] = {"unit": pulp_search.filters}

        if limit:
            body["criteria"]["limit"] = limit

        LOG.debug("Submitting %s unassociate: %s", url, body)

        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _get_repo_lock_data(self, repo_id):
        repo_raw_f = self._request_executor.submit(
            self._do_request,
            url=os.path.join(self._url, "pulp/api/v2/repositories/%s/" % repo_id),
            method="GET",
        )
        notes_f = f_map(repo_raw_f, lambda data: (data.get("notes") or {}))
        return f_map(
            notes_f,
            lambda notes: {key: notes[key] for key in notes if LOCK_CLAIM_STR in key},
        )

    def _update_repo_lock_data(self, repo_id, note_delta, await_result=False):
        update_f = self._request_executor.submit(
            self._do_request,
            url=os.path.join(self._url, "pulp/api/v2/repositories/%s/" % repo_id),
            method="PUT",
            json={"delta": {"notes": note_delta}},
        )
        if await_result:
            update_f.result()

    def _compile_notes(self, repo):
        # Given a repo we're about to publish, calculate and set certain notes
        # derived from the repo contents.
        out = f_return()

        if isinstance(repo, FileRepository):
            out = compile_ud_mappings(
                repo,
                # UD mapping code is written to do low-level pulp requests.
                # In order to keep the code at arms length and don't let it access
                # client internals directly, we'll just pass a callback
                # it can use to do these requests.
                do_request=lambda url, **kwargs: self._request_executor.submit(
                    self._do_request, url=os.path.join(self._url, url), **kwargs
                ),
            )

        return out

    @classmethod
    def _unpack_response(cls, pulp_response):
        try:
            parsed = pulp_response.json()
        except Exception:
            # Couldn't parse as JSON?
            # If the response was unsuccessful, raise that.
            # Otherwise re-raise parse error.
            pulp_response.raise_for_status()
            raise

        if (
            isinstance(parsed, dict)
            and parsed.get("http_status") == 404
            and parsed.get("http_request_method") == "DELETE"
        ):
            # Special case allowing unsuccessful status:
            # If we asked to DELETE something, and we got a 404 response,
            # this is not considered an error because the postcondition
            # of our request is satisfied.
            # Hence, don't call raise_for_status.
            pass
        else:
            # In general, we'll raise if response was unsuccessful.
            pulp_response.raise_for_status()

        return parsed

    @classmethod
    def _log_spawned_tasks(cls, taskdata):
        try:
            spawned = taskdata.get("spawned_tasks") or []
            for task in spawned:
                LOG.info("Created Pulp task: %s", task["task_id"])
        except Exception:  # pylint: disable=broad-except
            # something wrong with the data, can't log.
            # This error will be raised elsewhere
            pass
        return taskdata

    def _handle_page(self, urls, object_class, searches, raw_data_list):

        page_data = []
        for url, search, raw_data in zip(urls, searches, raw_data_list):
            # Extract metadata from Pulp units
            LOG.debug("Got pulp response for %s, %s elems", search, len(raw_data))
            if object_class is Unit and url.endswith("units/"):
                raw_data = [elem["metadata"] for elem in raw_data]

            page_data.extend([object_class.from_data(elem) for elem in raw_data])

        for obj in page_data:
            # set_client is only applicable for repository and distributor objects
            if hasattr(obj, "_set_client"):
                obj._set_client(self)

        # Do we need a next page?
        next_page = None

        next_responses = []
        next_searches = []
        next_urls = []
        for url, search, raw_data in zip(urls, searches, raw_data_list):
            limit = search["criteria"]["limit"]
            if len(raw_data) == limit:
                # Yeah, we might...
                search = search.copy()
                search["criteria"] = search["criteria"].copy()
                search["criteria"]["skip"] = search["criteria"]["skip"] + limit
                response_f = self._do_search(url, search)
                next_searches.append(search)
                next_responses.append(response_f)
                next_urls.append(url)

        if next_responses:
            next_page = f_proxy(
                f_map(
                    f_sequence(next_responses),
                    lambda data: self._handle_page(
                        next_urls, object_class, next_searches, data
                    ),
                )
            )
        return Page(data=page_data, next=next_page)

    @property
    def _session(self):
        if not hasattr(self._tls, "session"):
            self._tls.session = self._new_session()
        return self._tls.session

    def _new_session(self):
        out = requests.Session()
        for key, value in self._session_kwargs.items():
            # Special case for headers update as requests module already populates some defaults
            if key == "headers":
                out.headers.update(value)
            else:
                setattr(out, key, value)
        return out

    def _do_request(self, **kwargs):
        response = self._session.request(**kwargs)

        # Responses normally hold a reference back to their request.
        # We don't want to do that because our request might point to
        # large 'bytes' objects (when uploading files) and we want to allow
        # that to be reclaimed ASAP. If we hold onto this and our caller
        # has any cycles in play (and in practice they do), we'd be keeping
        # large chunks of memory pinned until potentially several GC
        # collections manage to run.
        response.request = None

        return response

    def _do_search(self, url, search):
        LOG.debug("Submitting %s search: %s", url, search)
        return self._request_executor.submit(
            self._do_request, method="POST", url=url, json=search
        )

    def _delete_resource(self, resource_type, resource_id):
        url = os.path.join(
            self._url, "pulp/api/v2/%s/%s/" % (resource_type, resource_id)
        )

        LOG.debug("Queuing request to DELETE %s", url)
        return self._task_executor.submit(self._do_request, method="DELETE", url=url)

    def _request_upload(self, name):
        url = os.path.join(self._url, "pulp/api/v2/content/uploads/")

        LOG.debug("Requesting upload id for %s", name)
        return self._request_executor.submit(self._do_request, method="POST", url=url)

    def _do_upload(self, data, upload_id, offset):
        url = os.path.join(
            self._url, "pulp/api/v2/content/uploads/%s/%s/" % (upload_id, offset)
        )

        return self._request_executor.submit(
            self._do_request, method="PUT", url=url, data=data
        )

    def _do_import(
        self, repo_id, upload_id, unit_type_id, unit_key, unit_metadata=None
    ):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/import_upload/" % repo_id
        )

        unit_metadata = unit_metadata or {}

        body = {
            "unit_type_id": unit_type_id,
            "upload_id": upload_id,
            "unit_key": unit_key,
            "unit_metadata": unit_metadata,
        }

        LOG.debug("Importing contents to repo %s with upload id %s", repo_id, upload_id)
        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _delete_upload_request(self, upload_id, name):
        url = os.path.join(self._url, "pulp/api/v2/content/uploads/%s/" % upload_id)

        LOG.debug("Deleting upload request %s for %s", upload_id, name)
        return self._request_executor.submit(self._do_request, method="DELETE", url=url)

    def _do_get_maintenance(self):
        def map_404_to_none(exception):
            # Translates 404 errors to a None response (no maintenance report).
            if (
                getattr(exception, "response", None) is not None
                and exception.response.status_code == 404
            ):
                return None
            # Any other types of errors are raised unchanged.
            raise exception

        url = os.path.join(self._url, "pulp/isos/redhat-maintenance/repos.json")

        response = self._request_executor.submit(
            self._do_request, method="GET", url=url
        )

        return f_map(response, error_fn=map_404_to_none)

    def _do_sync(self, repo_id, sync_options):
        if not sync_options["feed"]:
            raise ValueError("Cannot sync with empty feed: '%s'" % sync_options["feed"])

        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/sync/" % repo_id
        )

        body = {"override_config": sync_options}

        LOG.debug("Syncing repository %s with feed %s", repo_id, sync_options["feed"])
        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def create_repository(self, repo):
        """Create a repository with initial data provided in the
           argument. Importer for repository is automatically associated
           if available. If repository already exists, warning is logged.
           After repository has been successfully created or if repository already exists,
           it is checked if expected and current repository configuration values are
           correct.

        Args:
            repo (:class:`~pubtools.pulplib.Repository`)
                A repository object used for creation.

        Returns:
            Future[Repository]
                A future which is resolved with a value of ``Repository`` once the
                repository has been created.

        .. versionadded:: 2.39.0
        """
        url = os.path.join(self._url, "pulp/api/v2/repositories/")

        body = repo._to_data()
        repo_id = body["id"]

        importer = body.pop("importers", [])
        body["importer_type_id"] = importer[0]["importer_type_id"] if importer else None
        body["importer_config"] = importer[0]["config"] if importer else None

        # re-key distributor dict keys due to pulp API inconsistency
        for item in body["distributors"]:
            item["distributor_id"] = item["id"]
            item["distributor_config"] = item["config"]
            del item["config"]
            del item["id"]

        def log_existing_repo(exception):
            if (
                getattr(exception, "response", None) is not None
                and exception.response.status_code == 409
            ):
                LOG.warning("Repository %s already exists", repo_id)
                return None

            raise exception

        def check_repo(repo_on_server):
            # evolve some fields that don't have to be set before repository creation
            # but they're typically set automatically after creation call which results in
            # inequality between `repo_on_server` and `repo`
            dists = [evolve(item, repo_id=repo_id) for item in repo.distributors]
            repo_updated = evolve(
                repo, relative_url=repo_on_server.relative_url, distributors=dists
            )
            try:
                assert (
                    repo_on_server == repo_updated
                ), "Repo exists on server with unexpected values"
            except AssertionError:
                if importer:
                    self._check_resource(
                        "importer",
                        ["type_id", "config"],
                        repo_updated.importer,
                        repo_on_server.importer,
                        repo_updated.id,
                    )

                extra_dist = set()
                missing_dist = set()
                for expected_item in repo_updated.distributors:
                    for current_item in repo_on_server.distributors:
                        if expected_item.type_id == current_item.type_id:
                            self._check_resource(
                                "distributor",
                                ["id", "type_id", "last_publish"],
                                expected_item,
                                current_item,
                                repo_updated.id,
                            )
                            extra_dist.discard(current_item.type_id)
                            missing_dist.discard(expected_item.type_id)
                        else:
                            extra_dist.add(current_item.type_id)
                            missing_dist.add(expected_item.type_id)

                for item in extra_dist:
                    LOG.error(
                        "Repository %s contains unexpected distributor with type: %s",
                        repo_id,
                        item,
                    )
                for item in missing_dist:
                    LOG.error(
                        "Repository %s is missing distributor with type: %s",
                        repo_id,
                        item,
                    )

                LOG.exception(
                    "Repository %s exists on server and contains unexpected values",
                    repo_id,
                )
                raise

            return f_return(repo_on_server)

        LOG.debug("Creating repository %s", repo_id)
        out = self._request_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

        out = f_map(out, error_fn=log_existing_repo)
        out = f_flat_map(out, lambda _: self.get_repository(repo_id))
        out = f_flat_map(out, check_repo)

        return f_proxy(out)

    @staticmethod
    def _check_resource(
        resource_type, fields, expected_resource, current_resource, repo_id
    ):
        for attr in fields:
            expected = getattr(expected_resource, attr)
            current = getattr(current_resource, attr)
            if expected != current:
                LOG.error(
                    "Repository %s contains wrong %s %s\n"
                    "\t expected: %s\n"
                    "\t current: %s\n",
                    repo_id,
                    resource_type,
                    attr,
                    expected,
                    current,
                )



import os
import logging
import datetime
import threading

from pubtools.pulplib._impl.model import Task
from .errors import MissingTaskException, TaskFailedException


LOG = logging.getLogger("pubtools.pulplib")

# need to catch exceptions and pass onto descriptors in several places
# pylint: disable=broad-except


def task_log(task):
    parts = [task.id]
    for tag in task.tags or []:
        # All tags start with 'pulp:' prefix, which makes it useless noise
        # in the logs, so filter them out to save some horizontal space
        if tag.startswith("pulp:"):
            tag = tag[len("pulp:") :]
        parts.append(tag)
    return ", ".join(parts)


class TaskPoller(object):
    # Poll function used with PollExecutor.
    # Takes care of polling for Pulp task completion and resolving futures.

    # max number of times polling is attempted before errors are considered fatal
    MAX_ATTEMPTS = 60

    # delay in seconds between poll attempts
    DELAY = 5.0

    # delay in seconds between "Still waiting" log messages, if Pulp tasks
    # apparently are not being processed
    ACTIVITY_DELAY = datetime.timedelta(minutes=5)

    def __init__(self, session, url, timer=datetime.datetime.utcnow):
        self._session = session
        # Lock protects 'session', as cancel() may be called from one thread
        # while another is in the middle of poll, both using the session
        self.lock = threading.Lock()
        self.url = url
        self.attempt = 1
        self.timer = timer
        self.last_activity = self.timer()

    @property
    def session(self):
        # A helper to ensure all access to session is protected by lock.
        # If you try to use the requests.Session without the lock held,
        # it's a bug.
        assert self.lock.locked(), "INTERNAL ERROR: unsynchronized access to session"
        return self._session

    def __call__(self, descriptors):
        try:
            # Find every referenced Pulp task
            descriptor_task_ids, all_task_ids = self.gather_descriptor_tasks(
                descriptors
            )

            # Get status of all of those tasks from Pulp
            with self.lock:
                tasks = self.search_tasks(all_task_ids)

            # Now check all descriptors and decide which have completed
            resolved_count = self.resolve_descriptors(tasks, descriptor_task_ids)

            # Any successful poll resets the retry counter
            self.attempt = 1

            # We have activity if we resolved at least one thing (or if there was
            # nothing to do at all)
            if resolved_count or not descriptors:
                self.last_activity = self.timer()

            # If nothing's going on for a while, log about it
            with self.lock:
                self.log_if_inactive()
        except Exception:
            if self.attempt >= self.MAX_ATTEMPTS:
                LOG.exception("Pulp task polling repeatedly failed")
                raise

            LOG.warning(
                "Error occurred during Pulp task polling, will retry %d more time(s)",
                (self.MAX_ATTEMPTS - self.attempt),
                exc_info=1,
            )

            self.attempt += 1

        return self.DELAY

    def resolve_descriptors(self, tasks, descriptor_task_ids):
        resolved = 0
        for descriptor, task_ids in descriptor_task_ids:
            if self.resolve_descriptor(tasks, descriptor, task_ids):
                resolved += 1
        return resolved

    def resolve_descriptor(self, tasks, descriptor, task_ids):
        out = []

        for task_id in task_ids:
            task = tasks.get(task_id)

            if not task:
                # A task has been lost somehow
                exception = MissingTaskException(
                    "Task %s disappeared from Pulp!" % task_id
                )
                LOG.warning("%s", exception)
                descriptor.yield_exception(exception)
                return True

            if task.completed and not task.succeeded:
                LOG.warning("Pulp task failed: %s", task_log(task))

                exception = TaskFailedException(task)
                descriptor.yield_exception(exception)
                return True

            out.append(task)

        for task in out:
            if not task.completed:
                # can't resolve the future yet since there's a pending task
                return False

            LOG.info("Pulp task completed: %s", task_log(task))

        # OK, future can be resolved with the completed tasks
        descriptor.yield_result(out)
        return True

    def search_tasks(self, task_ids):
        out = {}

        if not task_ids:
            return out

        search_url = os.path.join(self.url, "pulp/api/v2/tasks/search/")
        search = {"criteria": {"filters": {"task_id": {"$in": task_ids}}}}
        LOG.debug("Searching %s task(s) at %s", len(task_ids), search_url)
        LOG.debug("Search criteria: %s", search)

        response = self.session.post(search_url, json=search)

        try:
            response.raise_for_status()
        except Exception as ex:
            # raise_for_status exception info is a little terse.
            # Try to log response body as well if we can, but be prepared
            # for that to fail as well.
            try:
                LOG.warning("Pulp task search failure: %s", response.json())
            except Exception:
                pass
            raise ex

        for elem in response.json():
            task = Task.from_data(elem)
            out[task.id] = task

        return out

    def gather_descriptor_tasks(self, descriptors):
        descriptor_tasks = []
        all_tasks = []

        for descriptor in descriptors:
            result = descriptor.result
            try:
                tasks = result.get("spawned_tasks") or []
                task_ids = [t["task_id"] for t in tasks]
                descriptor_tasks.append((descriptor, task_ids))
                all_tasks.extend(task_ids)
            except Exception as ex:
                LOG.warning("Not a valid call report from Pulp: %s", result, exc_info=1)
                descriptor.yield_exception(ex)

        return descriptor_tasks, all_tasks

    def cancel(self, task_data):
        # Invoked when a request is made to cancel a future.
        task_ids = [t["task_id"] for t in task_data["spawned_tasks"]]

        with self.lock:
            for task_id in task_ids:
                url = os.path.join(self.url, "pulp/api/v2/tasks/%s/" % task_id)
                response = self.session.delete(url)
                response.raise_for_status()

                LOG.info("Cancelled Pulp task: %s", task_id)

        return True

    def log_if_inactive(self):
        now = self.timer()
        delta = now - self.last_activity

        if delta < self.ACTIVITY_DELAY:
            # nothing to do (yet)
            return

        # We've been idle for too long, log a message to confirm we're not dead
        search_url = os.path.join(self.url, "pulp/api/v2/tasks/search/")
        search = {
            "criteria": {"filters": {"state": {"$in": ["running", "waiting"]}}},
            "fields": ["state"],
        }

        response = self.session.post(search_url, json=search)
        response.raise_for_status()
        tasks = response.json()

        running_count = len([t for t in tasks if t["state"] == "running"])
        waiting_count = len([t for t in tasks if t["state"] == "waiting"])

        LOG.info(
            "Still waiting on Pulp, load: %s running, %s waiting",
            running_count,
            waiting_count,
            extra={
                "event": {
                    "type": "awaiting-pulp",
                    "running-tasks": running_count,
                    "waiting-tasks": waiting_count,
                }
            },
        )

        # Logging that message counts as "activity"
        self.last_activity = now



import hashlib
import tempfile
import logging
import random
import threading
import uuid

import yaml
import attr

from pubtools.pulplib import (
    FileUnit,
    RpmUnit,
    RpmDependency,
    ErratumUnit,
    YumRepoMetadataFileUnit,
    ModulemdUnit,
    ModulemdDependency,
    ModulemdDefaultsUnit,
)

from .rpmlib import (
    get_rpm_header,
    get_header_fields,
    get_keys_from_header,
    get_rpm_requires,
    get_rpm_provides,
)
from ..model.attr import PULP2_UNIT_KEY

LOG = logging.getLogger("pubtools.pulplib")


class UnitMaker(object):
    def __init__(self, seen_unit_ids):
        self._lock = threading.RLock()
        self._uuidgen = random.Random()
        self._uuidgen.seed(0)
        self._seen_unit_ids = seen_unit_ids

    def next_unit_id(self):
        with self._lock:
            while True:
                next_raw_id = self._uuidgen.randint(0, 2**128)
                if next_raw_id not in self._seen_unit_ids:
                    break

        return str(uuid.UUID(int=next_raw_id))

    def make_units(self, type_id, unit_key, unit_metadata, content, repo_id):
        # Obtain valid unit(s) representing uploaded 'content'.
        if type_id == "iso":
            return [self.make_iso_unit(unit_key, unit_metadata)]

        if type_id == "yum_repo_metadata_file":
            return [self.make_yum_repo_metadata_unit(unit_key, unit_metadata)]

        if type_id == "rpm":
            return [self.make_rpm_unit(content, unit_metadata)]

        if type_id == "erratum":
            return [self.make_erratum_unit(unit_metadata)]

        if type_id == "modulemd":
            return self.make_module_units(content, repo_id)

        # Comps-related types are accepted, but we do not actually process
        # them into units.
        if type_id in (
            "package_group",
            "package_langpacks",
            "package_category",
            "package_environment",
        ):
            return []

        # It should not be possible to get here via public API.
        #
        # If you see this message, you're probably halfway through implementing
        # something. Well done, you've found the next place you need to update!
        raise NotImplementedError(
            "fake client does not support upload of '%s'" % type_id
        )  # pragma: no cover

    def make_iso_unit(self, unit_key, unit_metadata):
        usermeta = (unit_metadata or {}).get("pulp_user_metadata") or {}

        return FileUnit(
            unit_id=self.next_unit_id(),
            path=unit_key["name"],
            size=unit_key["size"],
            sha256sum=unit_key["checksum"],
            **usermeta
        )

    def make_yum_repo_metadata_unit(self, unit_key, unit_metadata):
        # For expected fields in unit_key vs metadata, refer to:
        # https://github.com/pulp/pulp_rpm/blob/5c5a7dcc058b29d89b3a913d29cfcab41db96686/plugins/pulp_rpm/plugins/importers/yum/upload.py#L246
        return YumRepoMetadataFileUnit(
            unit_id=self.next_unit_id(),
            data_type=unit_key["data_type"],
            sha256sum=unit_metadata["checksum"],
        )

    def make_erratum_unit(self, metadata):
        md = metadata.copy()
        md["_id"] = self.next_unit_id()
        md["_content_type_id"] = "erratum"
        return ErratumUnit.from_data(md)

    def make_rpm_unit(self, content, unit_metadata):
        # Since the native RPM library is used under the hood here,
        # any old file-like object isn't good enough; it *must* be a real file with
        # a 'fileno' which can be passed into syscalls. So we have to pipe the content
        # into a temporary file before we call into kobo.rpmlib.
        with tempfile.NamedTemporaryFile(suffix="pubtools-pulplib-fake") as f:
            # While we're at it, we can calculate each of the checksum types expected to
            # exist on the unit.
            hashers = {
                "md5sum": hashlib.md5(),  # nosec
                "sha1sum": hashlib.sha1(),  # nosec
                "sha256sum": hashlib.sha256(),
            }

            while True:
                chunk = content.read(1024 * 1024)
                if not chunk:
                    break
                f.write(chunk)
                for hasher in hashers.values():
                    hasher.update(chunk)

            sumattrs = {}
            for (sumtype, hasher) in hashers.items():
                sumattrs[sumtype] = hasher.hexdigest()

            # Read all the other attrs out of the RPM header.
            f.flush()
            f.seek(0)
            header = get_rpm_header(f)

        rpmattrs = get_header_fields(
            header, ["name", "version", "release", "arch", "epochnum", "sourcerpm"]
        )

        # Fill in some special cases:
        # This one uses a different name in the unit vs header.
        rpmattrs["epoch"] = str(rpmattrs.pop("epochnum"))

        # This one is derived from multiple headers.
        rpmattrs["signing_key"] = get_keys_from_header(header)
        if rpmattrs["signing_key"]:
            rpmattrs["signing_key"] = rpmattrs["signing_key"].lower()

        # This isn't actually a property of the RPM at all, but something synthesized.
        rpmattrs["filename"] = "{name}-{version}-{release}.{arch}.rpm".format(
            **rpmattrs
        )

        rpmattrs["requires"] = [
            RpmDependency._from_data(item) for item in get_rpm_requires(header)
        ]
        rpmattrs["provides"] = [
            RpmDependency._from_data(item) for item in get_rpm_provides(header)
        ]

        rpmattrs.update(sumattrs)

        usermeta = (unit_metadata or {}).get("pulp_user_metadata") or {}
        rpmattrs.update(usermeta)

        return RpmUnit(unit_id=self.next_unit_id(), **rpmattrs)

    def make_module_units(self, content, repo_id):
        # Note: consider using libmodulemd here once there is no need for this
        # code to support legacy (pre-RHEL8) environments.
        docs = list(yaml.load_all(content, Loader=yaml.BaseLoader))

        out = []
        for doc in docs:
            attrs = {}

            doc_type = doc.get("document")
            data = doc.get("data") or {}

            if doc_type == "modulemd":
                klass = ModulemdUnit

                # Pulp does not store entire 'artifacts' but only the 'rpms'
                # subkey
                artifacts = data.pop("artifacts", None)
                if artifacts and "rpms" in artifacts:
                    attrs["artifacts"] = artifacts["rpms"]

                dependencies = data.pop("dependencies", None)

                # Pulp does not store entire 'dependencies' but only the 'requires'
                # part, a little bit of parsing is required in the manner similar to pulp
                if dependencies:
                    parsed_deps = []
                    for dep in dependencies:
                        dep_item = {}
                        for k, v in dep["requires"].items():
                            dep_item[k] = v
                        parsed_deps.append(dep_item)

                    attrs["dependencies"] = ModulemdDependency._from_data(parsed_deps)

            elif doc_type == "modulemd-defaults":
                klass = ModulemdDefaultsUnit

                # This is always initialized using the repo for which we're uploading.
                attrs["repo_id"] = repo_id

                # For whatever reason, pulp renames this field, so it won't
                # be picked up by the generic attrs<=>unit mapping below.
                if "module" in data:
                    attrs["name"] = data["module"]

            else:
                # Pulp silently ignores any unknown document types, so we do the same.
                LOG.debug(
                    "FakeClient ignoring modulemd document of unknown type '%s'",
                    doc_type,
                )
                continue

            # attr names match the names used in modulemd data, so they can
            # simply be copied across.
            fields = [f.name for f in attr.fields(klass)]
            for field in fields:
                if field in data:
                    attrs[field] = data[field]

            # Note: if any mandatory fields are missing, or conversions can't
            # happen, this is where we crash.
            out.append(klass(unit_id=self.next_unit_id(), **attrs))

        return out


def is_erratum_version_newer(old_erratum, new_erratum):
    # A comparison of 'version' field between two erratum units which
    # aims to be compatible with:
    # https://github.com/pulp/pulp_rpm/blob/5c5a7dcc058b29d89b3a913d29cfcab41db96686/plugins/pulp_rpm/yum_plugin/util.py#L196

    old_version = getattr(old_erratum, "version", "")
    new_version = getattr(new_erratum, "version", "")

    if not new_version:
        return False

    try:
        new = float(new_version)

        if not old_version:
            old = 0
        else:
            old = float(old_version)
    except ValueError:
        # non-numeric field means we just treat new_version as newer.
        return True

    return new > old


def merge_units(old_unit, new_unit):
    # Given a unit which already exists in (fake) Pulp and a proposed new
    # unit, returns the new unit which should be stored - potentially
    # merging certain fields with the old unit.

    if old_unit is None:
        # No previous value in pulp
        return new_unit

    repos = None

    # If there's an existing unit, and either old or new unit has some
    # repo memberships, the repo memberships are merged.
    if (
        old_unit.repository_memberships is not None
        or new_unit.repository_memberships is not None
    ):
        repos = set(
            (old_unit.repository_memberships or [])
            + (new_unit.repository_memberships or [])
        )

    # By default, the new unit is used completely and it overwrites any
    # prior unit with the same key (excepting the repository_memberships).
    #
    # There is one special case for erratum units.
    # The new unit is only used if certain fields differ,
    # mainly 'version'. The actual logic is quite a bit more complicated
    # (refer to https://github.com/pulp/pulp_rpm/blob/5c5a7dcc058b29d89b3a913d29cfcab41db96686/plugins/pulp_rpm/plugins/db/models.py#L1233)
    # but our tools only use 'version' field for this, so the other complexity
    # seems not worth reproducing in the fake.
    if old_unit.content_type_id == "erratum" and not is_erratum_version_newer(
        old_unit, new_unit
    ):
        LOG.debug(
            "FakeClient ignoring erratum upload %s due to non-newer version",
            new_unit.id,
        )
        # Upload request is effectively ignored, the old unit is reused,
        # discarding any of the changed fields.
        #
        # Note: real Pulp also has quite complicated logic around 'pkglist'
        # field, we do not currently attempt to reproduce this in the fake.
        new_unit = old_unit

    return attr.evolve(new_unit, unit_id=old_unit.unit_id, repository_memberships=repos)


def make_unit_key(unit):
    # Given a unit, returns a unit key tuple used internally by the fake client.
    #
    # The unit keys returned here do not have to be precisely identical to those
    # used by a real Pulp server. However, they should be conceptually equivalent.
    # The fake client will keep all references to units via these keys, so the
    # fields used within a key controls whether an upload operation would add a new
    # unit or overwrite an existing one.
    #
    # Example: in Pulp it is not possible to store multiple ISO units with the same
    # (path, checksum, size). The fake client needs to use a unit key incorporating
    # those same fields, otherwise the fake could create situations which are
    # impossible on a real Pulp server.

    if isinstance(unit, FileUnit):
        return (unit.path, unit.sha256sum, unit.size)

    if isinstance(unit, YumRepoMetadataFileUnit):
        # This should not be possible because the unit_key embeds a repo_id.
        # If this somehow occurs, we can't continue.
        assert (
            len(unit.repository_memberships) == 1
        ), "BUG: a YumRepoMetadataFileUnit has been created without any repository!"
        return (unit.data_type, unit.repository_memberships[0])

    if isinstance(unit, RpmUnit):
        return (
            unit.name,
            unit.epoch,
            unit.version,
            unit.release,
            unit.arch,
            unit.sha256sum,
        )

    if isinstance(unit, ErratumUnit):
        return (unit.id,)

    if isinstance(unit, ModulemdUnit):
        return (unit.nsvca,)

    if isinstance(unit, ModulemdDefaultsUnit):
        return (unit.name, unit.repo_id)

    # Can't get here normally.
    # If you get here, you're probably partway through implementing a new type of unit.
    raise NotImplementedError(
        "fake client does not support '%s'" % unit
    )  # pragma: no cover


def with_key_only(units):
    # Given some units, returns copies of those units with only unit_key
    # fields present; mimics the behavior of pulp when filling units_successful
    # field on tasks.
    out = []
    for unit in units:
        klass = type(unit)
        kwargs = {}
        for field in attr.fields(klass):
            if field.metadata.get(PULP2_UNIT_KEY):
                kwargs[field.name] = getattr(unit, field.name)

        out.append(klass(**kwargs))

    return out


def with_filtered_fields(unit, fields):
    # Given a unit and a set of fields (e.g. from a prepared search),
    # returns a copy of that unit with only the specified fields (plus
    # a few special cases), similarly to a Pulp server's handling of the
    # 'fields' argument during a search.
    if fields is None:
        return unit

    model_field_names = set([f.model_field_name for f in fields])

    # These are special, not subject to field filtering.
    model_field_names.add("repository_memberships")
    model_field_names.add("unit_id")
    model_field_names.add("content_type_id")

    klass = type(unit)
    kwargs = {}
    for field in attr.fields(klass):
        if field.name in model_field_names:
            kwargs[field.name] = getattr(unit, field.name)

    return klass(**kwargs)



import logging

try:
    from time import monotonic
except ImportError:  # pragma: no cover
    from monotonic import monotonic


class TimedLogger(object):
    """A helper to log messages only if a certain amount of time has passed
    since the previous log.

    Intended for loops where you want to log progress sometimes but you don't
    know how fast the loop will run, making it potentially too spammy to
    log at every iteration or even at some fixed interval of iterations.
    """

    def __init__(self, logger=None, interval=10):
        """Construct a new timed logger.

        Arguments:
            logger
                Underlying logger object to which messages will be routed.

            interval
                Minimum amount of time, in seconds, which must pass between
                messages. Each time a message is logged, if the amount of
                time passed since the previous log is less than this, the
                message will be discarded.
        """
        logger = logger or logging.getLogger("pubtools.pulplib")

        self._interval = interval

        # We start counting from the time we're constructed. That means we
        # won't produce any log message at all until we've been alive for at
        # least 'interval' seconds.
        self._last_log = monotonic()

        # The log methods we support. Feel free to add others like warn/error
        # if needed, but it's not clear there's a use-case.
        self.debug = self._wrap(logger.debug)
        self.info = self._wrap(logger.info)

    def _wrap(self, logmethod):
        def new_logmethod(*args, **kwargs):
            now = monotonic()
            if now - self._last_log < self._interval:
                # Too soon, don't log yet
                return
            # OK, we should log
            self._last_log = now
            return logmethod(*args, **kwargs)

        return new_logmethod



from .base import Unit, type_ids_for_class
from .file import FileUnit
from .rpm import RpmUnit, RpmDependency
from .erratum import (
    ErratumUnit,
    ErratumReference,
    ErratumPackageCollection,
    ErratumPackage,
    ErratumModule,
)
from .repo_metadata import YumRepoMetadataFileUnit
from .modulemd import ModulemdUnit, ModulemdDependency
from .modulemd_defaults import ModulemdDefaultsUnit

SUPPORTED_UNIT_TYPES = ("iso", "rpm", "srpm", "modulemd", "modulemd_defaults")



import datetime

from .base import Unit, unit_type

from ..attr import pulp_attrib
from ... import compat_attr as attr
from ..convert import (
    frozenlist_or_none_sorted_converter,
    tolerant_timestamp,
    timestamp_converter,
)
from ..validate import optional_str, instance_of


@unit_type("iso")
@attr.s(kw_only=True, frozen=True)
class FileUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing a generic file.

    .. versionadded:: 1.5.0
    """

    path = pulp_attrib(type=str, pulp_field="name", unit_key=True)
    """Full path for this file in the Pulp repository.

    This may include leading directory components, but
    in the common case, this is only a basename.
    """

    size = pulp_attrib(type=int, converter=int, pulp_field="size", unit_key=True)
    """Size of this file, in bytes."""

    sha256sum = pulp_attrib(
        type=str,
        pulp_field="checksum",
        converter=lambda s: s.lower() if s else s,
        unit_key=True,
    )
    """SHA256 checksum of this file, as a hex string."""

    description = pulp_attrib(
        type=str,
        pulp_field="pulp_user_metadata.description",
        default=None,
        validator=optional_str,
    )
    """A user-oriented terse description of this file.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    version = pulp_attrib(
        type=str,
        pulp_field="pulp_user_metadata.version",
        default=None,
        validator=optional_str,
    )
    """A version string associated with this file.

    This string can be used for display purposes to group
    related files together. For example, files ``"oc-4.8.10-linux.tar.gz"``
    and ``"openshift-install-linux-4.8.10.tar.gz"`` in a repo may both
    have a version string of ``"4.8.10"`` to indicate that the files relate
    to the same product version.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.23.0
    """

    display_order = pulp_attrib(
        type=float,
        pulp_field="pulp_user_metadata.display_order",
        default=None,
        converter=lambda x: float(x) if x is not None else None,
    )
    """An ordering hint associated with this file.

    In cases where a UI displays a list of files from Pulp, it is suggested
    that files by default should be ordered by this field ascending.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.23.0
    """

    @display_order.validator
    def _validate_display_order(self, _, value):
        # Validation here is identical to pushsource.FilePushItem.display_order.
        # Check that validator for an explanation.
        if value is None:
            return

        value = float(value)

        if not (value >= -99999 and value <= 99999):
            raise ValueError("display_order must be within range -99999 .. 99999")

    cdn_path = pulp_attrib(
        type=str,
        pulp_field="pulp_user_metadata.cdn_path",
        default=None,
        validator=optional_str,
    )
    """A path, relative to the CDN root, from which this file can be downloaded
    once published.

    This path will not point to any specific repository. However, the file must
    have been published via at least one repository before this path can be
    accessed.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    cdn_published = pulp_attrib(
        type=datetime.datetime,
        pulp_field="pulp_user_metadata.cdn_published",
        default=None,
        converter=tolerant_timestamp,
        py_pulp_converter=timestamp_converter,
        validator=instance_of((datetime.datetime, type(None))),
    )
    """Approximate :class:`~datetime.datetime` in UTC at which this file first
    became available at ``cdn_path``, or ``None`` if this information is
    unavailable.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    content_type_id = pulp_attrib(
        default="iso", type=str, pulp_field="_content_type_id"
    )

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.

    .. versionadded:: 2.6.0
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """

    @size.validator
    def _check_size(self, _, value):
        if value < 0:
            raise ValueError("Not a valid size (must be positive): %s" % value)

    @sha256sum.validator
    def _check_sha256(self, _, value):
        self._check_sum(value, "SHA256", 64)



from .. import compat_attr as attr

optional = attr.validators.optional
instance_of = attr.validators.instance_of


@attr.s(kw_only=True, frozen=True, repr=True)
class CopyOptions(object):
    """Options influencing a call to
    :meth:`~pubtools.pulplib.Client.copy_content`.
    """

    require_signed_rpms = attr.ib(
        type=bool, default=None, validator=optional(instance_of(bool))
    )
    """Whether to require signatures on all RPMs in the copy.

    In order to copy unsigned RPMs between repositories, it is generally
    necessary to set this flag to ``False``. Unsigned RPMs may otherwise
    be silently omitted from the copy.
    """



import attr

ATTR_VERSION = tuple(int(x) for x in attr.__version__.split(".")[0:2])

# pylint: disable=invalid-name

# Wrappers around attr module to handle some incompatibilities with older versions of the module.


def s(*args, **kwargs):
    # A few features we'd like to use are not available on older attrs
    # versions, so we just silently disable them.
    #
    # Note this means we're technically exporting a different API on
    # older versions of attr, i.e. keywords-only isn't enforced.
    # This isn't ideal, but is mitigated by the fact that any new code
    # using this library will surely not be tested solely using old versions of attr.
    kwargs = kwargs.copy()

    if "slots" not in kwargs:
        # Slotted classes should be used where possible, but there
        # are a few cases where it won't work, so the possibility
        # is left open to disable it via slots=False.
        kwargs["slots"] = True

    if "repr" not in kwargs:
        # We provide our own __repr__ implementation in PulpObject, so
        # don't use the attrs-generated repr by default
        kwargs["repr"] = False

    if "kw_only" in kwargs and ATTR_VERSION < (18, 2):  # pragma: no cover
        # This is only implemented in attrs 18.2 and newer.
        # attr.s() will raise if kw_only is provided in older version of attr.
        del kwargs["kw_only"]

    return attr.s(*args, **kwargs)


if ATTR_VERSION < (18, 1):  # pragma: no cover
    # older attrs version didn't provide this function yet,
    # but it's easily added here.
    def fields_dict(cls):
        out = {}
        for field in attr.fields(cls):
            out[field.name] = field
        return out

else:
    fields_dict = attr.fields_dict


validators = attr.validators

if ATTR_VERSION < (19, 1):  # pragma: no cover
    # Backport is_callable, deep_iterable and deep_mapping methods needed elsewhere
    # for use with older attrs version
    @s(repr=False, slots=False, hash=True)
    class _IsCallableValidator:
        def __call__(self, _inst, attr_, value):
            if not callable(value):
                message = (
                    "'{name}' must be callable (got {value!r} that is a {actual!r})."
                )
                raise TypeError(
                    msg=message.format(
                        name=attr_.name, value=value, actual=value.__class__
                    ),
                    value=value,
                )

        def __repr__(self):
            return "<is_callable validator>"

    def is_callable():
        return _IsCallableValidator()

    @s(repr=False, slots=True, hash=True)
    class _DeepIterable:
        member_validator = attr.ib(validator=is_callable())
        iterable_validator = attr.ib(
            default=None, validator=validators.optional(is_callable())
        )

        def __call__(self, inst, attr_, value):
            # pylint: disable=not-callable
            if self.iterable_validator is not None:
                self.iterable_validator(inst, attr_, value)

            for member in value:
                self.member_validator(inst, attr_, member)
            # pylint: enable=not-callable

        def __repr__(self):
            iterable_identifier = (
                ""
                if self.iterable_validator is None
                else f" {self.iterable_validator!r}"
            )
            return (
                f"<deep_iterable validator for{iterable_identifier}"
                f" iterables of {self.member_validator!r}>"
            )

    def deep_iterable(member_validator, iterable_validator=None):
        if isinstance(member_validator, (list, tuple)):
            member_validator = validators.and_(*member_validator)
        return _DeepIterable(member_validator, iterable_validator)

    @s(repr=False, slots=True, hash=True)
    class _DeepMapping:
        key_validator = attr.ib(validator=is_callable())
        value_validator = attr.ib(validator=is_callable())
        mapping_validator = attr.ib(
            default=None, validator=validators.optional(is_callable())
        )

        def __call__(self, inst, attr_, value):
            # pylint: disable=not-callable
            if self.mapping_validator is not None:
                self.mapping_validator(inst, attr_, value)

            for key in value:
                self.key_validator(inst, attr_, key)
                self.value_validator(inst, attr_, value[key])
            # pylint: enable=not-callable

        def __repr__(self):
            return (
                f"<deep_mapping validator for objects mapping {self.key_validator!r} "
                f"to {self.value_validator!r}>"
            )

    def deep_mapping(key_validator, value_validator, mapping_validator=None):
        return _DeepMapping(key_validator, value_validator, mapping_validator)

    validators.is_callable = is_callable
    validators.deep_iterable = deep_iterable
    validators.deep_mapping = deep_mapping


ib = attr.ib
Factory = attr.Factory
fields = attr.fields
evolve = attr.evolve
has = attr.has
NOTHING = attr.NOTHING



try:
    from kobo.rpmlib import (
        get_rpm_header,
        get_header_fields,
        get_keys_from_header,
        parse_evr,
    )
except Exception as ex:  # pragma: no cover, pylint: disable=broad-except
    # Avoid a hard dependency on RPM bindings at import time - delay the crash
    # until we're actually used (if ever).
    #
    # Why: because there's no officially supported method of getting RPM bindings
    # in place by "pip install", making the dependency tricky on some environments.
    # rpmdyn can often be used, so we provide that hint as we crash,
    # but it's not appropriate to unconditionally depend on that.
    exception = ex

    def broken(*_args, **_kwargs):
        raise RuntimeError(
            "kobo.rpmlib is not available\nHint: consider 'pip install rpmdyn'"
        ) from exception

    get_rpm_header = broken
    get_header_fields = broken
    get_keys_from_header = broken
    parse_evr = broken

__all__ = [
    "get_rpm_header",
    "get_header_fields",
    "get_keys_from_header",
    "get_rpm_requires",
    "get_rpm_provides",
]


def get_rpm_requires(header):
    header_flags_map = {
        "version": "REQUIREVERSION",
        "name": "REQUIRENAME",
        "flags": "REQUIREFLAGS",
    }
    return _get_rpm_deps(header, header_flags_map)


def get_rpm_provides(header):
    header_flags_map = {
        "version": "PROVIDEVERSION",
        "name": "PROVIDENAME",
        "flags": "PROVIDEFLAGS",
    }
    return _get_rpm_deps(header, header_flags_map)


def _get_rpm_deps(header, header_flags_map):
    header_flags = list(header_flags_map.values())
    raw_deps = get_header_fields(header, header_flags)

    names = raw_deps[header_flags_map["name"]]
    versions = raw_deps[header_flags_map["version"]]
    flags = raw_deps[header_flags_map["flags"]]

    deps = []

    for name, evr, flag in zip(names, versions, flags):
        evr = parse_evr(evr, allow_empty_release=True) if evr else {}

        deps_item = {
            "name": name,
            "version": evr.get("version") or None,
            "release": evr.get("release") or None,
            "epoch": "0" if evr.get("epoch") == "" else None,
            "flags": _parse_dep_relation(flag),
        }

        deps.append(deps_item)

    return deps


def _parse_dep_relation(flag):
    LT = 0x02
    GT = 0x04
    EQ = 0x08

    if flag & LT and flag & EQ:
        flag_str = "LE"
    elif flag & GT and flag & EQ:
        flag_str = "GE"
    elif flag & LT:
        flag_str = "LT"
    elif flag & GT:
        flag_str = "GT"
    elif flag & EQ:
        flag_str = "EQ"
    else:
        flag_str = None

    return flag_str



import logging
import datetime
import contextlib
from pubtools.pulplib._impl.criteria import (
    AndCriteria,
    OrCriteria,
    FieldMatchCriteria,
    TrueCriteria,
    RegexMatcher,
    EqMatcher,
    InMatcher,
    ExistsMatcher,
    LessThanMatcher,
    UnitTypeMatchCriteria,
)

from pubtools.pulplib._impl import compat_attr as attr
from pubtools.pulplib._impl.model.attr import PULP2_FIELD, PY_PULP2_CONVERTER
from pubtools.pulplib._impl.model.unit.base import Unit, class_for_type_id
from pubtools.pulplib._impl.client.errors import AmbiguousQueryException

LOG = logging.getLogger("pubtools.pulplib")


def all_subclasses(klass):
    out = set()
    out.add(klass)
    for subclass in klass.__subclasses__():
        out.update(all_subclasses(subclass))
    return out


def to_mongo_json(value):
    # Return a value converted to the format expected for a mongo JSON
    # expression. Only a handful of special types need explicit conversions.
    if isinstance(value, datetime.datetime):
        return {"$date": value.strftime("%Y-%m-%dT%H:%M:%SZ")}

    if isinstance(value, (list, tuple)):
        return [to_mongo_json(elem) for elem in value]

    if isinstance(value, dict):
        out = {}
        for (key, val) in value.items():
            out[key] = to_mongo_json(val)
        return out

    return value


def raise_ambiguous_query(field_name, type_hint, grouped_class_names):
    # Raise an exception with a meaningful message in the case where a query
    # can't be executed due to field ambiguity.
    #
    # Example: if searching on field 'version', the caller might be intending
    # to use FileUnit.version or ModulemdUnit.version.
    # This message will point out the ambiguity and suggest how to resolve
    # the problem.

    message = ["Field '%s' is ambiguous." % field_name]
    message.append(
        "A subtype of %s must be provided to select one of the following groups:"
        % type_hint.__name__
    )
    field_lines = []
    for class_names in grouped_class_names:
        field_names = ["%s.%s" % (name, field_name) for name in sorted(class_names)]
        field_lines.append("  " + ", ".join(field_names))
    field_lines.sort()
    message.extend(field_lines)
    raise AmbiguousQueryException("\n".join(message))


def map_field_for_type(field_name, matcher, type_hint):
    if not type_hint:
        return None

    attrs_classes = all_subclasses(type_hint)
    attrs_classes = [cls for cls in attrs_classes if attr.has(cls)]
    found = {}
    for klass in attrs_classes:
        # Does the class have this field?
        klass_fields = attr.fields(klass)
        if not hasattr(klass_fields, field_name):
            continue
        field = getattr(klass_fields, field_name)
        metadata = field.metadata
        if PULP2_FIELD in metadata:
            pulp_field_name = metadata[PULP2_FIELD]
            converter = metadata.get(PY_PULP2_CONVERTER, lambda x: x)
            mapped = matcher._map(converter) if matcher else None
            key = (pulp_field_name, mapped)
            found.setdefault(key, []).append(klass.__name__)
            continue

        # Field was found on the model, but we don't support mapping it to
        # a Pulp field.
        raise NotImplementedError("Searching on field %s is not supported" % field_name)

    if len(found) == 1:
        # Found exactly one definition => great, use it
        return list(found.keys())[0]

    if len(found) > 1:
        # Found multiple definitions => a problem, we don't know what to use.
        return raise_ambiguous_query(field_name, type_hint, found.values())

    # No match => no change, search exactly what was requested
    return None


@attr.s
class PulpSearch(object):
    # Helper class representing a prepared Pulp search.
    # Usually just a filters dict, but may include type_ids
    # and fields for unit searches.
    filters = attr.ib(type=dict)
    type_ids = attr.ib(type=list, default=None)
    unit_fields = attr.ib(type=list, default=None)


class UnitTypeAccumulator(object):
    # A helper to accumulate info on the unit types we're dealing with
    # within a search:
    # - collects type_ids (the types we're searching for)
    # - collects unit_fields (the fields we want to query)
    # - complains if criteria is too complicated to gather a single consistent
    #   list of type_ids for the entire search
    #
    def __init__(self):
        self.can_accumulate = True
        self.type_ids = []

        # When no fields are requested, this must remain as None in order
        # to indicate that no field limits should be applied at all.
        self.unit_fields = None

    def accumulate_from_criteria(self, crit, match_expr):
        assert isinstance(crit, FieldMatchCriteria)

        # Accumulate type IDs here
        self.accumulate_from_match(match_expr)

        # And if this is specifically a criteria on unit type with
        # fields, accumulate those too
        if isinstance(crit, UnitTypeMatchCriteria) and crit._unit_fields is not None:
            self.unit_fields = self.unit_fields or set()
            self.unit_fields.update(crit._unit_fields)

    def accumulate_from_match(self, match_expr):
        # Are we still in a state where accumulation is possible?
        if not self.can_accumulate:
            raise ValueError(
                (
                    "Can't serialize criteria for Pulp query; too complicated. "
                    "Try simplifying the query with respect to unit_type/content_type_id."
                )
            )

        # OK, we can accumulate if it's a supported expression type.
        if isinstance(match_expr, dict) and list(match_expr.keys()) == ["$eq"]:
            self.type_ids = [match_expr["$eq"]]
        elif isinstance(match_expr, dict) and list(match_expr.keys()) == ["$in"]:
            self.type_ids = match_expr["$in"]
        else:
            raise ValueError(
                (
                    "Can't serialize criteria for Pulp query, "
                    "unsupported expression for content_type_id: %s\n"
                )
                % repr(match_expr)
            )

        # It is only possible to accumulate once.
        self.can_accumulate = False

    @property
    @contextlib.contextmanager
    def no_accumulate(self):
        old_can_accumulate = self.can_accumulate
        self.can_accumulate = False
        try:
            yield
        finally:
            self.can_accumulate = old_can_accumulate


def search_for_criteria(
    criteria, type_hint=None, unit_type_accum=None, accum_only=None
):
    # convert a Criteria object to a PulpSearch with filters as used in the Pulp 2.x API:
    # https://docs.pulpproject.org/dev-guide/conventions/criteria.html#search-criteria
    #
    # type_hint optionally provides the class expected to be found by this search.
    # This can impact the application of certain criteria, e.g. it will affect
    # field mappings looked up by FieldMatchCriteria.

    if type_hint is Unit and accum_only is None:
        # If the caller asked for a generic Unit type hint, we'll do two passes.
        # The first pass tries to figure out exactly what kind of Unit subtype is
        # being searched for. This is important because the same field might exist
        # on different models but be stored in Pulp in different ways (e.g.
        # FileUnit.version vs ModulemdUnit.version).
        accum = UnitTypeAccumulator()
        search_for_criteria(criteria, type_hint, accum, accum_only=True)

        # We should now know what type_ids we want to query. Map them back to
        # a specific unit type.
        unit_classes = set()
        for type_id in accum.type_ids:
            unit_class = class_for_type_id(type_id)
            if unit_class:
                unit_classes.add(unit_class)

        # If the type IDs map to exactly one unit subtype, then we update our
        # type_hint to be more accurate and continue.
        # Otherwise any field ambiguities should be detected later on
        # and a helpful error raised telling the caller to be more specific
        # with their query.
        if len(unit_classes) == 1:
            type_hint = list(unit_classes)[0]

    if criteria is None or isinstance(criteria, TrueCriteria):
        return PulpSearch(filters={})

    unit_type_accum = unit_type_accum or UnitTypeAccumulator()

    if isinstance(criteria, AndCriteria):
        clauses = [
            search_for_criteria(
                c, type_hint, unit_type_accum, accum_only=accum_only
            ).filters
            for c in criteria._operands
        ]

        # Empty filters do not affect the result and can be simplified.
        clauses = [c for c in clauses if c != {}]

        # A single clause is the same as no $and at all.
        if len(clauses) == 1:
            filters = clauses[0]
        else:
            filters = {"$and": clauses}

    elif isinstance(criteria, OrCriteria):
        with unit_type_accum.no_accumulate:
            filters = {
                "$or": [
                    search_for_criteria(
                        c, type_hint, unit_type_accum, accum_only=accum_only
                    ).filters
                    for c in criteria._operands
                ]
            }

    elif isinstance(criteria, FieldMatchCriteria):
        field = criteria._field
        matcher = criteria._matcher

        # Figure out the Pulp field being referenced here.
        #
        # If in accumulate-only mode (to calculate type_ids) we should skip this for now
        # because we may not have the final type_hint yet, which can lead to ambiguity.
        # content_type_id field itself is an exception as skipping that would interfere
        # with the calculation of type_ids.
        mapped = None
        if not accum_only or field == "content_type_id":
            mapped = map_field_for_type(field, matcher, type_hint)
        if mapped:
            field, matcher = mapped

        match_expr = field_match(matcher)

        # If we are looking at the special _content_type_id field
        # for the purpose of a unit search...
        if field == "_content_type_id" and issubclass(type_hint, Unit):
            # We should not include this into filters, but instead
            # attempt to accumulate it into type_ids_out.
            # This is because type_ids needs to be serialized into the
            # top-level 'criteria' and not 'filters', and there are
            # additional restrictions on its usage.
            unit_type_accum.accumulate_from_criteria(criteria, match_expr)

            filters = {}
        else:
            filters = {field: match_expr}

    else:
        raise TypeError("Not a criteria: %s" % repr(criteria))

    return PulpSearch(
        filters=filters,
        type_ids=unit_type_accum.type_ids[:],
        unit_fields=sorted(unit_type_accum.unit_fields)
        if unit_type_accum.unit_fields is not None
        else None,
    )


def filters_for_criteria(criteria, type_hint=None):
    return search_for_criteria(criteria, type_hint).filters


def field_match(to_match):
    if isinstance(to_match, RegexMatcher):
        return {"$regex": to_match._pattern}

    if isinstance(to_match, EqMatcher):
        return {"$eq": to_mongo_json(to_match._value)}

    if isinstance(to_match, InMatcher):
        return {"$in": to_mongo_json(to_match._values)}

    if isinstance(to_match, ExistsMatcher):
        return {"$exists": True}

    if isinstance(to_match, LessThanMatcher):
        return {"$lt": to_mongo_json(to_match._value)}

    raise TypeError("Not a matcher: %s" % repr(to_match))



from .base import Unit, unit_type

from ..attr import pulp_attrib
from ... import compat_attr as attr
from frozendict.core import frozendict  # pylint: disable=no-name-in-module
from ..convert import frozenlist_or_none_sorted_converter, frozendict_or_none_converter
from ..validate import optional_frozendict


@unit_type("modulemd_defaults")
@attr.s(kw_only=True, frozen=True)
class ModulemdDefaultsUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing a modulemd_defaults document.

    .. versionadded:: 2.4.0
    """

    name = pulp_attrib(type=str, pulp_field="name", unit_key=True)
    """The name of this modulemd defaults unit"""

    repo_id = pulp_attrib(type=str, pulp_field="repo_id", unit_key=True)
    """The repository ID bound to this modulemd defaults unit"""

    stream = pulp_attrib(type=str, pulp_field="stream", default=None)
    """The stream of this modulemd defaults unit"""

    profiles = pulp_attrib(
        type=frozendict,
        pulp_field="profiles",
        default=None,
        validator=optional_frozendict,
        converter=frozendict_or_none_converter,
    )
    """The profiles of this modulemd defaults unit."""

    content_type_id = pulp_attrib(
        default="modulemd_defaults", type=str, pulp_field="_content_type_id"
    )

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.

    .. versionadded:: 2.6.0
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """



import json
import random
import string
from time import sleep
import os
import logging

from datetime import datetime, timedelta
from more_executors.futures import f_map
from requests.exceptions import HTTPError

LOCK_CLAIM_STR = "pulplib-lock-claim-"

LOCK_VALID_FROM_OFFSET = int(os.getenv("PULPLIB_LOCK_VALID_FROM_OFFSET", "10"))
LOCK_EXPIRATION_OFFSET = int(os.getenv("PULPLIB_LOCK_EXPIRATION_OFFSET", "1800"))
LOCK_EXPIRATION_OFFSET_MIN = int(os.getenv("PULPLIB_LOCK_EXPIRATION_OFFSET_MIN", "300"))
LOCK_EXPIRATION_OFFSET_MAX = int(
    os.getenv("PULPLIB_LOCK_EXPIRATION_OFFSET_MAX", "216000")
)

AWAIT_ACTIVATION_SLEEP = int(os.getenv("PULPLIB_AWAIT_ACTIVATION_SLEEP", " 10"))

LOG = logging.getLogger("pubtools.pulplib")


# datetime is a built-in so cannot be mocked directly.
# This function is a workaround to allow mocking for testing.
def now():
    return datetime.now()  # pragma: no cover


class LockClaim(object):
    def __init__(self, **kwargs):
        expiration_offset = kwargs.get("expiration_offset")
        expiration_offset = (
            expiration_offset if expiration_offset else LOCK_EXPIRATION_OFFSET
        )
        expiration_offset = max(
            min(expiration_offset, LOCK_EXPIRATION_OFFSET_MAX),
            LOCK_EXPIRATION_OFFSET_MIN,
        )
        self.id = (
            kwargs["id"]
            if "id" in kwargs
            else "".join(random.choices(string.ascii_letters + string.digits, k=8))
        )
        self.context = kwargs["context"]
        self.created = (
            datetime.fromtimestamp(kwargs["created"]) if "created" in kwargs else now()
        )
        self.valid_from = (
            datetime.fromtimestamp(kwargs["valid_from"])
            if "valid_from" in kwargs
            else self.created + timedelta(seconds=LOCK_VALID_FROM_OFFSET)
        )
        self.expires = (
            datetime.fromtimestamp(kwargs["expires"])
            if "expires" in kwargs
            else self.created + timedelta(seconds=expiration_offset)
        )

    @property
    def as_json(self):
        data = {
            "id": self.id,
            "context": self.context,
            "created": datetime.timestamp(self.created),
            "valid_from": datetime.timestamp(self.valid_from),
            "expires": datetime.timestamp(self.expires),
        }
        return json.dumps(data)

    def __lt__(self, other):
        return self.created < other.created

    def __eq__(self, other):
        return self.id == other.id

    @staticmethod
    def from_json_data(json_data):
        return LockClaim(**json_data)

    @property
    def is_valid(self):
        return self.valid_from < now() < self.expires

    @property
    def is_expired(self):
        return now() > self.expires


class RepoLock(object):
    def __init__(
        self,
        repo_id,
        client,
        lock_context,
        expiration_offset,
    ):
        self._repo_id = repo_id
        self._client = client
        self._lock_claim = None
        self._lock_context = lock_context
        self._expiration_offset = expiration_offset

    def __enter__(self):
        LOG.info("Attempting to lock %s.", self._repo_id)
        self.remove_expired_locks()
        self._lock_claim = LockClaim(
            context=self._lock_context, expiration_offset=self._expiration_offset
        )
        self.submit_lock_claim()
        self.await_lock_activation()

    def __exit__(self, exc_type, exc_value, tb):
        if self._lock_claim.is_expired:
            logging.error(
                "Client requested lock on repo %s for %s seconds "
                "but was held for %s seconds (%s)",
                self._repo_id,
                (self._lock_claim.expires - self._lock_claim.created).seconds,
                (now() - self._lock_claim.created).seconds,
                self._lock_claim.context,
            )
        self.delete_lock_claim()
        self._lock_claim = None

    def get_repo_lock_data(self, filter_invalid=False):
        LOG.debug("Getting current repo lock data for %s.", self._repo_id)
        locks_f = self._client._get_repo_lock_data(self._repo_id)
        locks_f = f_map(
            locks_f,
            lambda locks: [
                LockClaim.from_json_data(json.loads(l)) for l in locks.values()
            ],
        )
        parsed_locks = locks_f.result()
        parsed_locks.sort()
        if filter_invalid:
            parsed_locks = [l for l in parsed_locks if l.is_valid]
        return parsed_locks

    def await_lock_activation(self):
        # if `now` is later than `valid_from`, the timedelta is negative and
        # the seconds value becomes a whole day minus the timedelta ~ 86400.
        # total_seconds returns a negative time value, which we clamp.
        sleep(max((self._lock_claim.valid_from - now()).total_seconds(), 0))
        while True:
            lock_claims = self.get_repo_lock_data(filter_invalid=True)
            # If our claim is the oldest valid lock claim.
            if self._lock_claim == lock_claims[0]:
                LOG.info("Lock obtained for %s.", self._repo_id)
                return
            LOG.info("Awaiting lock for %s.", self._repo_id)
            sleep(AWAIT_ACTIVATION_SLEEP)

    def submit_lock_claim(self):
        LOG.info(
            "Submitting lock with id '%s' to %s (%s).",
            self._lock_claim.id,
            self._repo_id,
            self._lock_claim.context,
        )
        self._client._update_repo_lock_data(
            self._repo_id,
            {LOCK_CLAIM_STR + self._lock_claim.id: self._lock_claim.as_json},
        )

    def delete_lock_claim(self):
        LOG.info(
            "Deleting lock with id '%s' from %s (%s).",
            self._lock_claim.id,
            self._repo_id,
            self._lock_claim.context,
        )
        note_delta = {LOCK_CLAIM_STR + self._lock_claim.id: None}
        try:
            self._client._update_repo_lock_data(
                self._repo_id, note_delta, await_result=True
            )
        except HTTPError as e:
            if self._pulp_exception_from_missing_lock(e, list(note_delta.keys())):
                logging.error("Lock %s has already been deleted.", self._lock_claim.id)
            else:
                raise e  # pragma: no cover

    def _pulp_exception_from_missing_lock(self, e, lock_names):
        # If we try to delete a lock from pulp notes that's already been
        # deleted, it will result in a 500 error due to a KeyError exception
        # If multiple locks are sent for deletion, it appears Pulp will fail on
        # the first missing one and not carry out any deletions.
        lock_names = ["KeyError: '%s'\n" % lock for lock in lock_names]
        return any(
            [exception in lock_names for exception in e.response.json()["exception"]]
        )

    def remove_expired_locks(self):
        expired_locks = [lock for lock in self.get_repo_lock_data() if lock.is_expired]
        if expired_locks:
            for lock in expired_locks:
                LOG.warning(
                    "Removing expired lock with id %s from %s (%s).",
                    lock.id,
                    self._repo_id,
                    lock.context,
                )
            invalid_locks_delta = {
                LOCK_CLAIM_STR + lock.id: None for lock in expired_locks
            }
            try:
                self._client._update_repo_lock_data(self._repo_id, invalid_locks_delta)
            except HTTPError as e:
                if self._pulp_exception_from_missing_lock(
                    e, list(invalid_locks_delta.keys())
                ):
                    logging.error(
                        "An error occurred while trying to delete an expired "
                        "lock. The locks have already been deleted"
                    )
                else:
                    raise e  # pragma: no cover



import datetime
import logging
import warnings
import json
from functools import partial

from attr import validators, asdict, converters

from frozenlist2 import frozenlist
from more_executors.futures import f_proxy, f_map, f_flat_map

from frozendict.core import frozendict  # pylint: disable=no-name-in-module
from .repo_lock import RepoLock
from ..attr import pulp_attrib, PULP2_FIELD, PULP2_MUTABLE
from ..common import PulpObject, Deletable, DetachedException
from ..convert import frozenlist_or_none_converter, frozendict_or_none_converter
from ..distributor import Distributor
from ...criteria import Criteria, Matcher
from ...schema import load_schema
from ... import compat_attr as attr
from ...hooks import pm
from ...util import dict_put, lookup, ABSENT

from ..common import schemaless_init

LOG = logging.getLogger("pubtools.pulplib")

REPO_CLASSES = {}


def repo_type(pulp_type):
    # decorator for Repository subclasses, registers against a
    # particular value of notes._repo-type
    def decorate(klass):
        REPO_CLASSES[pulp_type] = klass
        return klass

    return decorate


@attr.s(kw_only=True, frozen=True)
class Importer(PulpObject):
    """
    Importer is a pulp object that needs to be associated with repository
    in order to successfully sync or upload content to it.
    """

    type_id = pulp_attrib(default=None, type=str, pulp_field="importer_type_id")
    """
    Type id of the importer.
    """
    config = pulp_attrib(
        default=attr.Factory(frozendict),
        type=frozendict,
        converter=frozendict_or_none_converter,
        pulp_field="config",
    )
    """
    Configuration dictionary of the importer.
    """

    @classmethod
    def _from_data(cls, data):
        # Convert from raw list/dict as provided in Pulp responses into model.
        if isinstance(data, list):
            return cls._from_data(data[0]) if data else schemaless_init(cls, data)

        return schemaless_init(cls, data)

    def _to_data(self):
        return [
            {
                "importer_type_id": self.type_id,
                "config": self.config,
            }
        ]


@attr.s(kw_only=True, frozen=True)
class PublishOptions(object):
    """Options controlling a repository
    :meth:`~pubtools.pulplib.Repository.publish`.
    """

    force = pulp_attrib(default=None, type=bool)
    """If True, Pulp should publish all data within a repository, rather than attempting
    to publish only changed data (or even skipping the publish).

    Setting ``force=True`` may have a major performance impact when publishing large repos.
    """

    clean = pulp_attrib(default=None, type=bool)
    """If True, certain publish tasks will not only publish new/changed content, but
    will also attempt to erase formerly published content which is no longer present
    in the repo.

    Setting ``clean=True`` generally implies ``force=True``.
    """

    origin_only = pulp_attrib(default=None, type=bool)
    """If ``True``, Pulp should only update the content units / origin path on
    remote hosts.

    Only relevant if a repository has one or more distributors where
    :meth:`~pubtools.pulplib.Distributor.is_rsync` is ``True``.
    """

    rsync_extra_args = pulp_attrib(
        default=None, type=list, converter=frozenlist_or_none_converter
    )
    """If present, provide these additional arguments to any rsync commands run
    during publish.

    Ignored when rsync is not used.
    """


@attr.s(kw_only=True, frozen=True)
class SyncOptions(object):
    """Options controlling a repository
    :meth:`~pubtools.pulplib.Repository.sync`.

    .. seealso:: Subclasses for specific repository
                 types: :py:class:`~pubtools.pulplib.FileSyncOptions`,
                 :py:class:`~pubtools.pulplib.YumSyncOptions`,
                 :py:class:`~pubtools.pulplib.ContainerSyncOptions`
    """

    feed = pulp_attrib(type=str)
    """URL where the repository's content will be synchronized from.
    """

    ssl_validation = pulp_attrib(default=None, type=bool)
    """Indicates if the server's SSL certificate is verified against the CA certificate uploaded.
    """

    ssl_ca_cert = pulp_attrib(default=None, type=str)
    """CA certificate string used to validate the feed source's SSL certificate
    """

    ssl_client_cert = pulp_attrib(default=None, type=str)
    """Certificate used as the client certificate when synchronizing the repository
    """

    ssl_client_key = pulp_attrib(default=None, type=str)
    """Private key to the certificate specified in ssl_client_cert
    """

    max_speed = pulp_attrib(default=None, type=int)
    """The maximum download speed in bytes/sec for a task (such as a sync).

    Default is None
    """

    proxy_host = pulp_attrib(default=None, type=str)
    """A string representing the URL of the proxy server that should be used when synchronizing
    """

    proxy_port = pulp_attrib(default=None, type=int)
    """An integer representing the port that should be used when connecting to proxy_host.
    """

    proxy_username = pulp_attrib(default=None, type=str)
    """A string representing the username that should be used to authenticate with the proxy server
    """

    proxy_password = pulp_attrib(default=None, type=str)
    """A string representing the password that should be used to authenticate with the proxy server
    """

    basic_auth_username = pulp_attrib(default=None, type=str)
    """Username to authenticate with source which supports basic authentication.
    """

    basic_auth_password = pulp_attrib(default=None, type=str)
    """Password to authenticate with source which supports basic authentication.
    """


def pv_converter(versions):
    # Converter for values in a product_versions field.
    #
    # We try to sort numerically while decomposing dotted versions into
    # their components, so e.g. "8.10" is sorted later than "8.8".
    # However, we don't know for sure which version strings might be stored
    # in the field (or even non-strings), so we can fall back to a generic
    # string sort.

    # Everything is initially interpreted as a string regardless of how
    # it was stored.
    versions = [str(v) for v in versions]

    try:
        return sorted(
            versions, key=lambda version: [int(c) for c in str(version).split(".")]
        )
    except ValueError:
        return sorted(versions)


@attr.s(kw_only=True, frozen=True)
class Repository(PulpObject, Deletable):
    """Represents a Pulp repository."""

    _SCHEMA = load_schema("repository")

    # The distributors (by ID) which should be activated when publishing this repo.
    # Order matters. Distributors which don't exist will be ignored.
    _PUBLISH_DISTRIBUTORS = [
        "iso_distributor",
        "yum_distributor",
        "cdn_distributor",
        "cdn_distributor_unprotected",
        "docker_web_distributor_name_cli",
    ]

    id = pulp_attrib(type=str, pulp_field="id")
    """ID of this repository (str)."""

    type = pulp_attrib(default=None, type=str, pulp_field="notes._repo-type")
    """Type of this repository (str).

    This is a brief string denoting the content / Pulp plugin type used with
    this repository, e.g. ``rpm-repo``.
    """

    created = pulp_attrib(
        default=None, type=datetime.datetime, pulp_field="notes.created"
    )
    """:class:`~datetime.datetime` in UTC at which this repository was created,
    or None if this information is unavailable.
    """

    distributors = pulp_attrib(
        default=attr.Factory(frozenlist),
        type=list,
        pulp_field="distributors",
        converter=frozenlist,
        pulp_py_converter=lambda ds: frozenlist([Distributor.from_data(d) for d in ds]),
        # It's too noisy to let repr descend into sub-objects
        repr=False,
    )
    """list of :class:`~pubtools.pulplib.Distributor` objects belonging to this
    repository.
    """

    eng_product_id = pulp_attrib(
        default=None,
        type=int,
        pulp_field="notes.eng_product",
        pulp_py_converter=converters.optional(int),
        py_pulp_converter=converters.optional(str),
    )
    """ID of the product to which this repository belongs (if any)."""

    relative_url = pulp_attrib(default=None, type=str)
    """Default publish URL for this repository, relative to the Pulp content root."""

    mutable_urls = pulp_attrib(
        default=attr.Factory(frozenlist), type=list, converter=frozenlist
    )
    """A list of URLs relative to repository publish root which are expected
    to change at every publish (if any content of repo changed)."""

    is_sigstore = pulp_attrib(default=False, type=bool)
    """True if this is a sigstore repository, used for container image manifest
    signatures.

    .. deprecated:: 2.24.0
       The signatures are not stored in a Pulp repository any more.
    """

    is_temporary = pulp_attrib(
        default=False,
        type=bool,
        validator=validators.instance_of(bool),
        pulp_field="notes.pub_temp_repo",
    )
    """True if this is a temporary repository.

    A temporary repository is a repository created by release-engineering tools
    for temporary use during certain workflows.  Such repos are not expected to
    be published externally and generally should have a lifetime of a few days
    or less.

    .. versionadded:: 1.3.0
    """

    signing_keys = pulp_attrib(
        default=attr.Factory(frozenlist),
        type=list,
        pulp_field="notes.signatures",
        pulp_py_converter=lambda sigs: sigs.split(",") if sigs else [],
        py_pulp_converter=",".join,
        converter=lambda keys: frozenlist([k.strip() for k in keys]),
    )
    """A list of GPG signing key IDs used to sign content in this repository."""

    skip_rsync_repodata = pulp_attrib(default=False, type=bool)
    """True if this repository is explicitly configured such that a publish of
    this repository will not publish repository metadata to remote hosts.
    """

    content_set = pulp_attrib(default=None, type=str, pulp_field="notes.content_set")
    """Name of content set that is associated with this repository."""

    arch = pulp_attrib(default=None, type=str, pulp_field="notes.arch")
    """The primary architecture of content within this repository (e.g. 'x86_64').

    .. versionadded:: 2.29.0
    """

    platform_full_version = pulp_attrib(
        default=None, type=str, pulp_field="notes.platform_full_version"
    )
    """A version string associated with the repository.

    This field should be used with care, as the semantics are not well defined.
    It is often, but not always, equal to the $releasever yum variable associated
    with a repository.

    Due to the unclear meaning of this field, it's strongly recommended to avoid
    making use of it in any new code.

    .. versionadded:: 2.29.0
    """

    product_versions = pulp_attrib(
        default=None,
        type=list,
        pulp_field="notes.product_versions",
        pulp_py_converter=json.loads,
        py_pulp_converter=partial(json.dumps, separators=(",", ":")),
        converter=partial(frozenlist_or_none_converter, map_fn=pv_converter),
        mutable=True,
    )
    """A list of product versions associated with this repository.

    The versions found in this list are derived from the product versions found in any
    product certificates (productid) historically uploaded to this repository and
    related repositories.

    This field is **mutable** and may be set by :meth:`~Client.update_repository`.

    .. versionadded:: 2.29.0
    """

    include_in_download_service = pulp_attrib(
        default=False,
        type=bool,
        mutable=True,
        pulp_field="notes.include_in_download_service",
        pulp_py_converter=lambda x: x == "True",
        py_pulp_converter=str,
    )
    """Flag indicating whether the repository is visible in production instance
    of download service.

    .. versionadded:: 2.34.0
    """

    include_in_download_service_preview = pulp_attrib(
        default=False,
        type=bool,
        mutable=True,
        pulp_field="notes.include_in_download_service_preview",
        pulp_py_converter=lambda x: x == "True",
        py_pulp_converter=str,
    )
    """Flag indicating whether the repository is visible in staging instance
    of download service.

    .. versionadded:: 2.34.0
    """

    provisioning = pulp_attrib(default=None, type=bool, pulp_field="notes.provisioning")
    """Flag indicating whether the repository is currently in process of creation
    and provisioning to other tools outside of Pulp.

    .. versionadded:: 2.37.0
    """

    importer = pulp_attrib(
        default=Importer(),
        type=Importer,
        pulp_field="importers",
        pulp_py_converter=Importer._from_data,
        py_pulp_converter=Importer._to_data,
    )
    """
    An object of :class:`~pubtools.pulplib.Importer` that is associated with the repository.

    .. versionadded:: 2.39.0
    """

    @distributors.validator
    def _check_repo_id(self, _, value):
        # checks if distributor's repository id is same as the repository it
        # is attached to
        for distributor in value:
            if not distributor.repo_id:
                return
            if distributor.repo_id == self.id:
                return
            raise ValueError(
                "repo_id doesn't match for %s. repo_id: %s, distributor.repo_id: %s"
                % (distributor.id, self.id, distributor.repo_id)
            )

    @property
    def _distributors_by_id(self):
        out = {}
        for dist in self.distributors:
            out[dist.id] = dist
        return out

    @classmethod
    def _mutable_note_fields(cls):
        # Returns the subset of fields on this class which are stored under the
        # notes dict and considered mutable, and thus can potentially be updated.
        return [
            fld
            for fld in attr.fields(cls)
            if fld.metadata.get(PULP2_FIELD, "").startswith("notes.")
            and fld.metadata.get(PULP2_MUTABLE)
        ]

    @property
    def _mutable_notes(self):
        # Returns notes dict containing only mutable notes, appropriate
        # for updating the repo.

        # Get self in raw Pulp form.
        self_raw = self._to_data()

        # Make a filtered view keeping only mutable note fields.
        out = {}

        for field in self._mutable_note_fields():
            pulp_field = field.metadata.get(PULP2_FIELD)
            pulp_value = lookup(self_raw, pulp_field)

            if pulp_value is not ABSENT:
                dict_put(out, pulp_field, pulp_value)

        # Return only the notes portion.
        return out.get("notes") or {}

    def distributor(self, distributor_id):
        """Look up a distributor by ID.

        Returns:
            :class:`~pubtools.pulplib.Distributor`
                The distributor belonging to this repository with the given ID.
            None
                If this repository has no distributor with the given ID.
        """
        return self._distributors_by_id.get(distributor_id)

    @property
    def file_content(self):
        """A list of file units stored in this repository.

        Returns:
            list[:class:`~pubtools.pulplib.FileUnit`]

        .. versionadded:: 2.4.0
        """
        return list(self.search_content(Criteria.with_field("content_type_id", "iso")))

    @property
    def rpm_content(self):
        """A list of rpm units stored in this repository.

        Returns:
            list[:class:`~pubtools.pulplib.RpmUnit`]

        .. versionadded:: 2.4.0
        """
        return list(self.search_content(Criteria.with_field("content_type_id", "rpm")))

    @property
    def srpm_content(self):
        """A list of srpm units stored in this repository.

        Returns:
            list[:class:`~pubtools.pulplib.Unit`]

        .. versionadded:: 2.4.0
        """
        return list(self.search_content(Criteria.with_field("content_type_id", "srpm")))

    @property
    def modulemd_content(self):
        """A list of modulemd units stored in this repository.

        Returns:
            list[:class:`~pubtools.pulplib.ModulemdUnit`]

        .. versionadded:: 2.4.0
        """
        return list(
            self.search_content(Criteria.with_field("content_type_id", "modulemd"))
        )

    @property
    def modulemd_defaults_content(self):
        """A list of modulemd_defaults units stored in this repository.

        Returns:
            list[:class:`~pubtools.pulplib.ModulemdDefaultsUnit`]

        .. versionadded:: 2.4.0
        """
        return list(
            self.search_content(
                Criteria.with_field("content_type_id", "modulemd_defaults")
            )
        )

    def search_content(self, criteria=None):
        """Search this repository for content matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Unit` objects.

        .. versionadded:: 2.4.0
        """
        if not self._client:
            raise DetachedException()

        return self._client._search_repo_units(self.id, criteria)

    def delete(self):
        """Delete this repository from Pulp.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when the repository deletion has completed.

                The future contains a list of zero or more tasks triggered and awaited
                during the delete operation.

                This object also becomes detached from the client; no further updates
                are possible.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.
        """
        return self._delete("repositories", self.id)

    def publish(self, options=PublishOptions()):
        """Publish this repository.

        The specific operations triggered on Pulp in order to publish a repo are not defined,
        but in Pulp 2.x, generally consists of triggering one or more distributors in sequence.

        Args:
            options (PublishOptions)
                Options used to customize the behavior of this publish.

                If omitted, the Pulp server's defaults apply.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when publish succeeds.

                The future contains a list of zero or more tasks triggered and awaited
                during the publish operation.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.
        """
        if not self._client:
            raise DetachedException()

        # Before adding distributors and publishing, we'll activate this hook
        # to allow subscribing implementers the opportunity to adjust options.
        hook_rets = pm.hook.pulp_repository_pre_publish(
            repository=self, options=options
        )
        # Use the first non-None hook return value to replace options.
        hook_rets = [ret for ret in hook_rets if ret is not None]
        options = hook_rets[0] if hook_rets else options

        # All distributor IDs we're willing to invoke. Anything else is ignored.
        # They'll be invoked in the order listed here.
        candidate_distributor_ids = self._PUBLISH_DISTRIBUTORS

        to_publish = []

        for candidate in candidate_distributor_ids:
            distributor = self._distributors_by_id.get(candidate)
            if not distributor:
                # nothing to be done
                continue

            if (
                distributor.id == "docker_web_distributor_name_cli"
                and options.origin_only
            ):
                continue

            config = self._config_for_distributor(distributor, options)
            to_publish.append((distributor, config))

        out = self._client._publish_repository(self, to_publish)

        def do_published_hook(tasks):
            # Whenever we've published successfully, we'll activate this hook
            # before returning.
            pm.hook.pulp_repository_published(repository=self, options=options)
            return tasks

        out = f_map(out, do_published_hook)
        return f_proxy(out)

    def sync(self, options=None):
        """Sync repository with feed.

        Args:
            options (SyncOptions)
                Options used to customize the behavior of sync process.
                If omitted, the Pulp server's defaults apply.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when sync succeeds.

                The future contains a list of zero or more tasks triggered and awaited
                during the sync operation.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 2.5.0
        """
        options = options or SyncOptions(feed="")

        if not self._client:
            raise DetachedException()

        return f_proxy(
            self._client._do_sync(
                self.id, asdict(options, filter=lambda name, val: val is not None)
            )
        )

    def lock(self, context, duration=None):
        """
        Obtain an exclusive advisory lock on this repository.

        Returns a context manager representing the lock, intended to be used
        via a `with` statement. When the context is entered, the caller will
        wait until the lock can be acquired (or raise an exception if the lock
        can't be acquired).

        Only a single :class:`~pubtools.pulplib.Client` is able to hold the lock
        on a repository at any given time. The lock does not prevent modifications
        to the repo with the Pulp API, and does not affect other Pulp client
        implementations or instances of :class:`~pubtools.pulplib.Client` not
        using the `lock` method.

        Args:
            context:
                A short description of the task being carried out with the lock.

                This value will be added to the lock in the repo and may be
                used for debugging.

            duration
                Maximum duration of the lock, in seconds.

                This value is used only if this client fails to release the
                lock (for example, because the current process is killed).
                In this case, the duration will be used by other clients in
                order to detect and release stale locks, avoiding a deadlock.

                There is no way to extend the duration of an acquired lock,
                so the caller should always ensure they request a `duration`
                high enough to cover the entire expected lifetime of the lock.
        """

        return RepoLock(self.id, self._client, context, duration)

    def remove_content(self, criteria=None, **kwargs):
        """Remove all content of requested types from this repository.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used to filter the contents for removal.

                Type IDs must be included in the criteria with any other filters.
                If omitted, filter criteria will be ignored and all the content will
                be removed.
                If criteria is omitted, all the content will be removed.

            limit (None, int)
                Limit the maximum number of units that will be disassociated by
                pulp.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when content has been removed.

                The future contains a list of zero or more tasks triggered and awaited
                during the removal.

                To obtain information on the removed content, use
                :meth:`~pubtools.pulplib.Task.units`.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 1.5.0
        """
        if not self._client:
            raise DetachedException()

        # Type IDs are must for the criteria filter to be effective. It must be included
        # in the criteria. Type IDs provided as type_id kwargs will be ignored for the
        # criteria and will remove all the content in the repo.
        # Refer to https://bugzilla.redhat.com/show_bug.cgi?id=1021579 and Pulp
        # documentation for more details.

        # Note: type_ids is deprecated. Criteria.with_unit_type should be used to
        # filter on type_ids. This is kept for backward compatibility and will be
        # removed in future versions.
        if not criteria:
            type_ids = kwargs.get("type_ids")
            # use content_type_id field name to coerce
            # search_for_criteria to fill out the PulpSearch#type_ids field.
            # passing a criteria with an empty type_ids list rather than
            # None results in failing tests due to the implementation of
            # FakeClient#_do_unassociate
            if type_ids is not None:
                warnings.warn(
                    "type_ids is deprecated, use criteria instead", DeprecationWarning
                )
                criteria = Criteria.with_field(
                    "content_type_id",
                    Matcher.in_(type_ids),  # Criteria.with_field_in is deprecated
                )

        return f_proxy(
            self._client._do_unassociate(
                self.id,
                criteria=criteria,
                limit=kwargs.get("limit"),
            )
        )

    @classmethod
    def from_data(cls, data):
        # delegate to concrete subclass as needed
        if cls is Repository:
            notes = data.get("notes") or {}
            for notes_type, klass in REPO_CLASSES.items():
                if notes.get("_repo-type") == notes_type:
                    return klass.from_data(data)

        return super(Repository, cls).from_data(data)

    @classmethod
    def _data_to_init_args(cls, data):
        out = super(Repository, cls)._data_to_init_args(data)

        for dist in data.get("distributors") or []:
            if dist["distributor_type_id"] in ("yum_distributor", "iso_distributor"):
                out["relative_url"] = (dist.get("config") or {}).get("relative_url")

            if dist["id"] == "cdn_distributor":
                skip_repodata = (dist.get("config") or {}).get("skip_repodata")
                if skip_repodata is not None:
                    out["skip_rsync_repodata"] = skip_repodata

        return out

    @classmethod
    def _config_for_distributor(cls, distributor, options):
        out = {}

        if distributor.is_rsync:
            if options.clean is not None:
                out["delete"] = options.clean
            if options.origin_only is not None:
                out["content_units_only"] = options.origin_only
            if options.rsync_extra_args is not None:
                out["rsync_extra_args"] = options.rsync_extra_args

        if options.force is not None:
            out["force_full"] = options.force

        return out

    def _set_client(self, client):
        super(Repository, self)._set_client(client)

        # distributors use the same client as owning repository
        for distributor in self.distributors or []:
            distributor._set_client(client)

    def _upload_then_import(
        self, file_obj, name, type_id, unit_key_fn=None, unit_metadata_fn=None
    ):
        """Private helper to upload and import a piece of content into this repo.

        To be called by the type-specific subclasses (e.g. YumRepository,
        FileRepository...)

        Args:
            file_obj (str, file-like object, None):
                file object or path (as documented in public methods), or None
                if this unit type has no associated file

            name (str):
                a brief user-meaningful name for the content being uploaded
                (appears in logs)

            type_id (str):
                pulp unit type ID

            unit_key_fn (callable):
                a callable which will be invoked with the return value of
                _do_upload_file (or None if file_obj is None).
                It should return the unit key for this piece of
                content. If omitted, an empty unit key is used, which means Pulp
                is wholly responsible for calculating the unit key.

            unit_metadata_fn (callable):
                a callable which will be invoked with the return value of
                _do_upload_file (or None if file_obj is None). It should return
                the unit metadata for this piece of
                content. If omitted, metadata is not included in the import call to
                Pulp.
        """

        if not self._client:
            raise DetachedException()

        unit_key_fn = unit_key_fn or (lambda _: {})
        unit_metadata_fn = unit_metadata_fn or (lambda _: None)

        upload_id_f = f_map(
            self._client._request_upload(name), lambda upload: upload["upload_id"]
        )

        f_map(
            upload_id_f,
            lambda upload_id: LOG.info(
                "Uploading %s to %s [%s]", name, self.id, upload_id
            ),
        )

        if file_obj is None:
            # If there is no file for this kind of unit (e.g. erratum),
            # we still have to use the request_upload and import APIs; we just
            # never upload any bytes. That means the upload is 'complete' as
            # soon as the upload ID is known. A real upload returns a (size, checksum)
            # tuple; we force a no-content upload to return None.
            upload_complete_f = f_map(upload_id_f, lambda _: None)
        else:
            upload_complete_f = f_flat_map(
                upload_id_f,
                lambda upload_id: self._client._do_upload_file(
                    upload_id, file_obj, name
                ),
            )

        import_complete_f = f_flat_map(
            upload_complete_f,
            lambda upload: self._client._do_import(
                self.id,
                upload_id_f.result(),
                type_id,
                unit_key_fn(upload),
                unit_metadata_fn(upload),
            ),
        )

        f_map(
            import_complete_f,
            lambda _: self._client._delete_upload_request(upload_id_f.result(), name),
        )

        return f_proxy(import_complete_f)



import datetime

from .common import PulpObject, Deletable, DetachedException
from .attr import pulp_attrib
from ..schema import load_schema
from .. import compat_attr as attr


@attr.s(kw_only=True, frozen=True)
class Distributor(PulpObject, Deletable):
    """Represents a Pulp distributor."""

    _SCHEMA = load_schema("repository", "distributor")

    id = pulp_attrib(type=str, pulp_field="id")
    """ID of this distributor (str).

    This is an arbitrary string, though often matches exactly the `type_id`.
    """

    type_id = pulp_attrib(type=str, pulp_field="distributor_type_id")
    """Type ID of this distributor (str).

    The type ID of a distributor determines which content may be handled and
    which steps may be performed by the distributor. For example, distributors
    of type `yum_distributor` may be used to create yum repositories.
    """

    repo_id = pulp_attrib(type=str, default=None, pulp_field="repo_id")
    """The :class:`pubtools.pulplib.Repository` ID this distributor is attached to."""

    relative_url = pulp_attrib(type=str, default=None, pulp_field="config.relative_url")
    """Default distribution URL for the repository which the distributor is attached to,
    relative to the Pulp content root."""

    last_publish = pulp_attrib(
        default=None, type=datetime.datetime, pulp_field="last_publish"
    )
    """The :class:`~datetime.datetime` at which this distributor was last published,
    if known."""

    is_rsync = attr.ib(
        default=attr.Factory(
            lambda self: self.type_id
            in (
                "rpm_rsync_distributor",
                "iso_rsync_distributor",
                "docker_rsync_distributor",
            ),
            takes_self=True,
        )
    )
    """True for distributors in the 'rsync distributor' family
    (e.g. ``rpm_rsync_distributor``).
    """

    def delete(self):
        """Delete this distributor from Pulp.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when the distributor deletion has completed.

                The future contains a list of zero or more tasks triggered and awaited
                during the delete operation.

                This object also becomes detached from the client; no further updates
                are possible.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client or repository.

        .. versionadded:: 2.3.0
        """
        if not self.repo_id:
            raise DetachedException()

        return self._delete("repositories/%s/distributors" % self.repo_id, self.id)



from .base import Repository, SyncOptions, repo_type
from ..attr import pulp_attrib
from ... import compat_attr as attr


@attr.s(kw_only=True, frozen=True)
class ContainerSyncOptions(SyncOptions):
    """Options controlling a container repository
    :meth:`~pubtools.pulplib.Repository.sync`.
    """

    upstream_name = pulp_attrib(default=None, type=str)
    """The name of the repository to import from the upstream repository.
    For example, if syncing from repository `quay.io/fedora/fedora`, upstream_name should be set to `fedora/fedora`.
    """

    tags = pulp_attrib(default=None, type=list)
    """List of tags to include on sync.
    """


@repo_type("docker-repo")
@attr.s(kw_only=True, frozen=True)
class ContainerImageRepository(Repository):
    """A :class:`~pubtools.pulplib.Repository` for container images."""

    type = pulp_attrib(default="docker-repo", type=str, pulp_field="notes._repo-type")

    registry_id = pulp_attrib(
        default=attr.Factory(lambda self: self.id, takes_self=True), type=str
    )
    """The ID of this repository in a container image registry.

    For example:

    - pulp repo id: redhat-rhel7-openscap
    - registry id:  rhel7/openscap

    The registry id is used by clients of the published repo, i.e.

    ``docker pull registry.example.com/<registry_id>:latest``
    """

    @classmethod
    def _data_to_init_args(cls, data):
        out = super(ContainerImageRepository, cls)._data_to_init_args(data)

        for dist in data.get("distributors") or []:
            if dist["distributor_type_id"] == "docker_distributor_web":
                registry_id = dist["config"].get("repo-registry-id")
                if registry_id:
                    out["registry_id"] = registry_id
                break

        return out



from .base import Unit, unit_type

from ..attr import pulp_attrib
from ... import compat_attr as attr
from ..convert import frozenlist_or_none_sorted_converter


@unit_type("yum_repo_metadata_file")
@attr.s(kw_only=True, frozen=True)
class YumRepoMetadataFileUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing a metadata file in a yum repo.

    .. versionadded:: 2.17.0
    """

    data_type = pulp_attrib(type=str, pulp_field="data_type", unit_key=True)
    """The type of this metadata file, e.g. "productid"."""

    sha256sum = pulp_attrib(
        type=str,
        pulp_field="checksum",
        converter=lambda s: s.lower() if s else s,
        default=None,
    )
    """SHA256 checksum of this metadata file, if known, as a hex string."""

    content_type_id = pulp_attrib(
        default="yum_repo_metadata_file", type=str, pulp_field="_content_type_id"
    )

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """



import random
import hashlib
import json
import re

from collections import namedtuple

from io import StringIO, BytesIO

from more_executors.futures import f_return, f_return_error, f_flat_map, f_proxy

from pubtools.pulplib import (
    Page,
    PulpException,
    Criteria,
    Task,
    Repository,
    Distributor,
    Unit,
    RpmUnit,
    MaintenanceReport,
    CopyOptions,
)
from pubtools.pulplib._impl.client.client import UploadResult
from pubtools.pulplib._impl.client.search import search_for_criteria
from .. import compat_attr as attr

from .match import match_object
from . import units

Publish = namedtuple("Publish", ["repository", "tasks"])
Upload = namedtuple("Upload", ["repository", "tasks", "name", "sha256"])
Sync = namedtuple("Sync", ["repository", "tasks", "sync_config"])
RepoLockRecord = namedtuple("RepoLockRecord", ["repository", "action"])


class FakeClient(object):  # pylint:disable = too-many-instance-attributes
    # Client implementation holding data in memory rather than
    # using a remote Pulp server.
    #
    # This class is not public, but it must have all of the same public
    # API as the pubtools.pulplib.Client class. The idea is that any code
    # written against pubtools.pulplib.Client should be able to work with
    # an instance of this class swapped in.
    _PAGE_SIZE = 3

    def __init__(self, state):
        self._state = state
        self._shutdown = False

    def __enter__(self):
        return self

    def __exit__(self, *_args, **_kwargs):
        self._shutdown = True

    def _ensure_alive(self):
        if self._shutdown:
            # We are technically capable of working just fine after shutdown,
            # but the point of this class is to be an accurate stand-in for
            # a real client, so raise the same kind of exception here
            raise RuntimeError("cannot schedule new futures after shutdown")

    def search_repository(self, criteria=None):
        self._ensure_alive()

        criteria = criteria or Criteria.true()
        repos = []

        # Pass the criteria through the code used by the real client to build
        # up the Pulp query. We don't actually *use* the resulting query since
        # we're not accessing a real Pulp server. The point is to ensure the
        # same validation and error behavior as used by the real client also
        # applies to the fake.
        search_for_criteria(criteria, Repository)

        try:
            for repo in self._state.repositories[:]:
                if match_object(criteria, repo):
                    repos.append(self._attach_repo(repo))
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        # callers should not make any assumption about the order of returned
        # values. Encourage that by returning output in unpredictable order
        random.shuffle(repos)
        return self._prepare_pages(repos)

    def search_content(self, criteria=None):
        self._ensure_alive()

        criteria = criteria or Criteria.true()
        out = []

        # Pass the criteria through the code used by the real client to build
        # up the Pulp query. We don't actually *use* the resulting query since
        # we're not accessing a real Pulp server. The point is to ensure the
        # same validation and error behavior as used by the real client also
        # applies to the fake.
        prepared_search = search_for_criteria(criteria, Unit)

        available_type_ids = set(self._state.type_ids)
        missing_type_ids = set(prepared_search.type_ids or []) - available_type_ids
        if missing_type_ids:
            return f_return_error(
                PulpException(
                    "following type ids are not supported on the server: %s"
                    % ",".join(missing_type_ids)
                )
            )

        for unit in self._state.all_units:
            if (
                prepared_search.type_ids
                and unit.content_type_id not in prepared_search.type_ids
            ):
                continue
            if match_object(criteria, unit):
                unit = units.with_filtered_fields(unit, prepared_search.unit_fields)
                out.append(unit)

        # callers should not make any assumption about the order of returned
        # values. Encourage that by returning output in unpredictable order
        random.shuffle(out)
        return self._prepare_pages(out)

    def copy_content(
        self, from_repository, to_repository, criteria=None, options=CopyOptions()
    ):
        self._ensure_alive()

        from_id = from_repository.id
        to_id = to_repository.id

        found = list(from_repository.search_content(criteria).result())

        # RPM signature filter: if signatures are required, unsigned RPMs are not
        # included in the copy.
        # Because we don't model this flag on distributor objects and because in
        # practice it's set to True, we default to True.
        if options.require_signed_rpms is not False:
            found = [u for u in found if not isinstance(u, RpmUnit) or u.signing_key]

        # Units are being copied to this repo, so that value obviously must appear
        # in repository_memberships from now on.
        found = [attr.evolve(unit, repository_memberships=[to_id]) for unit in found]

        with self._state.lock:
            # Now put the found units into the destination repo.
            # Any kind of merging or replacing of units is handled within this step.
            self._state.insert_repo_units(to_id, found)

            # Arbitrarily limit the number of units included per task. The point is
            # to enforce that the caller doesn't expect any specific number of tasks.
            tasks = []
            while found:
                next_batch = found[:5]
                found = found[5:]
                tasks.append(
                    Task(
                        id=self._state.next_task_id(),
                        repo_id=from_id,
                        completed=True,
                        succeeded=True,
                        units=units.with_key_only(next_batch),
                    )
                )

            if not tasks:
                # This indicates that nothing was found at all.
                # That's fine, just return a task with empty units.
                tasks.append(
                    Task(
                        id=self._state.next_task_id(),
                        repo_id=from_id,
                        completed=True,
                        succeeded=True,
                        units=[],
                    )
                )

        return f_proxy(f_return(tasks))

    def update_content(self, unit):
        self._ensure_alive()

        if not unit.unit_id:
            raise ValueError("unit_id missing on call to update_content()")

        with self._state.lock:
            # The unit has to exist.
            existing_unit = None
            for candidate in self._state.all_units:
                if (
                    candidate.content_type_id == unit.content_type_id
                    and candidate.unit_id == unit.unit_id
                ):
                    existing_unit = candidate
                    break
            else:
                return f_return_error(
                    PulpException("unit not found: %s" % unit.unit_id)
                )

            # OK, we have a unit to update. Figure out which fields we can update.
            update = {}
            for fld in unit._usermeta_fields():
                update[fld.name] = getattr(unit, fld.name)

            updated_unit = attr.evolve(existing_unit, **update)

            unit_key = units.make_unit_key(updated_unit)
            self._state.units_by_key[unit_key] = updated_unit

        return f_return()

    def update_repository(self, repository):
        self._ensure_alive()

        with self._state.lock:
            existing_repo = None
            for candidate in self._state.repositories:
                if candidate.id == repository.id:
                    existing_repo = candidate
                    break
            else:
                return f_return_error(
                    PulpException("repository not found: %s" % repository.id)
                )

            # We've got a repo, now update it.
            update = {}
            for fld in existing_repo._mutable_note_fields():
                update[fld.name] = getattr(repository, fld.name)

            updated_repo = attr.evolve(existing_repo, **update)

            self._state.repositories = [
                repo for repo in self._state.repositories if repo.id != updated_repo.id
            ] + [updated_repo]

        return f_return()

    def search_distributor(self, criteria=None):
        self._ensure_alive()

        criteria = criteria or Criteria.true()
        distributors = []

        search_for_criteria(criteria, Distributor)

        try:
            for repo in self._state.repositories[:]:
                for distributor in repo.distributors:
                    if match_object(criteria, distributor):
                        distributors.append(attr.evolve(distributor, repo_id=repo.id))
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        random.shuffle(distributors)
        return self._prepare_pages(distributors)

    def search_task(self, criteria=None):
        self._ensure_alive()
        tasks = []

        criteria = criteria or Criteria.true()
        search_for_criteria(criteria)

        try:
            for task in self._state.tasks[:]:
                if match_object(criteria, task):
                    tasks.append(task)
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        random.shuffle(tasks)
        return self._prepare_pages(tasks)

    def _search_repo_units(self, repo_id, criteria):
        criteria = criteria or Criteria.true()

        # Pass the criteria through the same handling as used by the real client
        # for serialization, to ensure we reject criteria also rejected by real client
        # and also accumulate unit_fields.
        prepared_search = search_for_criteria(criteria, Unit)

        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            return repo_f

        with self._state.lock:
            repo_units = self._state.repo_units(repo_id)

        out = []

        try:
            for unit in repo_units:
                if match_object(criteria, unit):
                    unit = units.with_filtered_fields(unit, prepared_search.unit_fields)
                    out.append(unit)
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        random.shuffle(out)
        return self._prepare_pages(out)

    def _prepare_pages(self, resource_list):
        # Split resource_list into pages
        # resource_list: list of objects that paginated
        page_data = []
        current_page_data = []
        while resource_list:
            next_elem = resource_list.pop()
            current_page_data.append(next_elem)
            if len(current_page_data) == self._PAGE_SIZE and resource_list:
                page_data.append(current_page_data)
                current_page_data = []

        page_data.append(current_page_data)

        page = Page()
        next_page = None
        for batch in reversed(page_data):
            page = Page(data=batch, next=next_page)
            next_page = f_proxy(f_return(page))

        return f_proxy(f_return(page))

    def get_repository(self, repository_id):
        if not isinstance(repository_id, str):
            raise TypeError("Invalid argument: id=%s" % id)

        data = self.search_repository(Criteria.with_id(repository_id)).result().data
        if len(data) != 1:
            return f_return_error(
                PulpException("Repository id=%s not found" % repository_id)
            )

        return f_proxy(f_return(data[0]))

    def get_maintenance_report(self):
        self._ensure_alive()

        with self._state.lock:
            if self._state.maintenance_report:
                report = MaintenanceReport._from_data(
                    json.loads(self._state.maintenance_report)
                )
            else:
                report = MaintenanceReport()
        return f_proxy(f_return(report))

    def set_maintenance(self, report):
        self._ensure_alive()

        report_json = json.dumps(report._export_dict(), indent=4, sort_keys=True)
        report_fileobj = StringIO(report_json)

        repo = self.get_repository("redhat-maintenance").result()

        # upload updated report to repository and publish
        upload_ft = repo.upload_file(report_fileobj, "repos.json")

        publish_ft = f_flat_map(upload_ft, lambda _: repo.publish())
        self._state.maintenance_report = report_json

        return f_proxy(publish_ft)

    def get_content_type_ids(self):
        self._ensure_alive()

        return f_proxy(f_return(self._state.type_ids))

    def _do_upload_file(
        self, upload_id, file_obj, name="<unknown file>"
    ):  # pylint: disable=unused-argument
        # We keep track of uploaded content as we may need it at import time.
        buffer = BytesIO()

        with self._state.lock:
            self._state.uploads_pending[upload_id] = buffer

        is_file_obj = "close" in dir(file_obj)
        if not is_file_obj:
            file_obj = open(file_obj, "rb")

        def do_next_upload(checksum, size):
            while True:
                data = file_obj.read(1024 * 1024)
                if not data:
                    break
                if isinstance(data, str):
                    data = data.encode("utf-8")
                buffer.write(data)
                checksum.update(data)
                size += len(data)

            return f_return(UploadResult(checksum.hexdigest(), size))

        out = f_flat_map(f_return(), lambda _: do_next_upload(hashlib.sha256(), 0))

        out.add_done_callback(lambda _: file_obj.close())

        return out

    def _do_unassociate(self, repo_id, criteria=None, limit=None):
        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            return repo_f

        with self._state.lock:
            current = self._state.repo_unit_keys.get(repo_id, set())
            units_with_key = [
                {"key": key, "unit": self._state.units_by_key[key]} for key in current
            ]
            removed_units = set()
            kept_keys = set()

            criteria = criteria or Criteria.true()
            # validating the criteria here like in actual scenario.
            pulp_search = search_for_criteria(
                criteria, type_hint=Unit, unit_type_accum=None
            )

            # raise an error if criteria with filters doesn't include type_ids
            if pulp_search.filters and not pulp_search.type_ids:
                raise ValueError(
                    "Criteria to remove_content must specify at least one unit type!"
                )

            for unit_with_key in units_with_key:
                unit = unit_with_key["unit"]
                if match_object(criteria, unit) and (
                    not limit or len(removed_units) < limit
                ):
                    removed_units.add(unit)
                else:
                    kept_keys.add(unit_with_key["key"])

            self._state.repo_unit_keys[repo_id] = kept_keys

            task = Task(
                id=self._state.next_task_id(),
                repo_id=repo_id,
                completed=True,
                succeeded=True,
                units=units.with_key_only(removed_units),
            )

        return f_return([task])

    def _get_repo_lock_data(self, repo_id):
        self._ensure_alive()
        data = self.search_repository(Criteria.with_id(repo_id)).result().data
        if len(data) != 1:
            return f_return_error(PulpException("Repository id=%s not found" % repo_id))
        with self._state.lock:
            data = (
                self._state.repo_locks[repo_id]
                if repo_id in self._state.repo_locks
                else {}
            )
            return f_return(data)

    # pylint: disable=unused-argument
    def _update_repo_lock_data(self, repo_id, note_delta, await_result=None):
        self._ensure_alive()
        data = self.search_repository(Criteria.with_id(repo_id)).result().data
        if len(data) != 1:
            return f_return_error(PulpException("Repository id=%s not found" % repo_id))
        with self._state.lock:

            data = (
                self._state.repo_locks[repo_id]
                if repo_id in self._state.repo_locks
                else {}
            )
            for lock_id in note_delta:
                data[lock_id] = note_delta[lock_id]

            data = {k: data[k] for k in data if data[k] is not None}
            self._state.repo_locks[repo_id] = data

            if len(note_delta) == 1:
                lock_id = list(note_delta)[0]
                self._state.repo_lock_history.append(
                    RepoLockRecord(repo_id, "lock" if note_delta[lock_id] else "unlock")
                )
            else:
                # The only time more than one change should be made is when
                # removing multiple expired locks.
                self._state.repo_lock_history.append(
                    RepoLockRecord(repo_id, "multi-unlock")
                )

    def _request_upload(self, name):  # pylint: disable=unused-argument
        # Note: old versions had a bug where this function would always
        # consume *two* request IDs. We keep that side-effect so that test
        # data produced with that bug remains stable.
        self._state.next_request_id()

        upload_id = self._state.next_request_id()
        upload_request = {
            "_href": "/pulp/api/v2/content/uploads/%s/" % upload_id,
            "upload_id": "%s" % upload_id,
        }

        return f_return(upload_request)

    def _do_import(
        self, repo_id, upload_id, unit_type_id, unit_key, unit_metadata=None
    ):
        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            # Repo can't be found, let that exception propagate
            return repo_f

        repo = repo_f.result()

        with self._state.lock:
            # Get the uploaded content we're about to import; though it's not
            # guaranteed to be present (e.g. erratum has no file).
            # If not present, we just use an empty BytesIO.
            upload_content = self._state.uploads_pending.pop(upload_id, BytesIO())
            upload_content.seek(0)

            new_units = self._state.unitmaker.make_units(
                unit_type_id, unit_key, unit_metadata, upload_content, repo_id
            )
            new_units = [
                attr.evolve(u, repository_memberships=[repo.id]) for u in new_units
            ]

            self._state.insert_repo_units(repo_id, new_units)

            task = Task(id=self._state.next_task_id(), completed=True, succeeded=True)

            # upload_history is a deprecated field, data is maintained for iso only.
            if unit_type_id == "iso":
                self._state.upload_history.append(
                    Upload(repo, [task], unit_key["name"], unit_key["checksum"])
                )

        return f_return([task])

    def _delete_resource(self, resource_type, resource_id):
        if resource_type == "repositories":
            return self._delete_repository(resource_id)

        match = re.match(r"^repositories/([^/]+)/distributors$", resource_type)
        if match:
            return self._delete_distributor(match.group(1), resource_id)

        # There is no way to get here using public API
        raise AssertionError(
            "Asked to delete unexpected %s" % resource_type
        )  # pragma: no cover

    def _delete_repository(self, repo_id):
        with self._state.lock:
            found = False
            for idx, repo in enumerate(self._state.repositories):
                if repo.id == repo_id:
                    found = True
                    break

            if not found:
                # Deleting something which already doesn't exist is fine
                return f_return([])

            self._state.repositories.pop(idx)  # pylint: disable=undefined-loop-variable
            self._state.repo_unit_keys.pop(repo_id, None)
            return f_return(
                [Task(id=self._state.next_task_id(), completed=True, succeeded=True)]
            )

    def _delete_distributor(self, repo_id, distributor_id):
        with self._state.lock:
            repo_f = self.get_repository(repo_id)
            if repo_f.exception():
                # Repo can't be found, let that exception propagate
                return repo_f

            repo = repo_f.result()
            new_distributors = [
                dist for dist in repo.distributors if dist.id != distributor_id
            ]
            dist_found = new_distributors != repo.distributors

            if not dist_found:
                # Deleting something which already doesn't exist is fine
                return f_return([])

            idx = self._state.repositories.index(repo)
            self._state.repositories[idx] = attr.evolve(
                repo, distributors=new_distributors
            )

            return f_return(
                [
                    Task(
                        id=self._state.next_task_id(),
                        completed=True,
                        succeeded=True,
                        tags=[
                            "pulp:repository:%s" % repo_id,
                            "pulp:repository_distributor:%s" % distributor_id,
                            "pulp:action:remove_distributor",
                        ],
                    )
                ]
            )

    def _publish_repository(self, repo, distributors_with_config):
        repo_f = self.get_repository(repo.id)
        if repo_f.exception():
            # Repo can't be found, let that exception propagate
            return repo_f

        with self._state.lock:
            tasks = []
            for _ in distributors_with_config:
                tasks.append(
                    Task(id=self._state.next_task_id(), completed=True, succeeded=True)
                )

            self._state.publish_history.append(Publish(repo, tasks))

        return f_return(tasks)

    def _attach_repo(self, repo):
        kwargs = {}
        if repo.distributors:
            # Deep copy for accurate attach/detach semantics
            kwargs["distributors"] = [attr.evolve(dist) for dist in repo.distributors]

        repo = attr.evolve(repo, **kwargs)
        repo._set_client(self)
        return repo

    def _do_sync(self, repo_id, sync_config):  # pylint:disable = unused-argument
        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            # Repo can't be found, let that exception propagate
            return repo_f

        with self._state.lock:
            task = Task(id=self._state.next_task_id(), completed=True, succeeded=True)

            self._state.sync_history.append(Sync(repo_f.result(), [task], sync_config))

        return f_return([task])

    def create_repository(self, repo):
        with self._state.lock:
            if repo.id not in [
                existing_repo.id for existing_repo in self._state.repositories
            ]:
                self._state.repositories.append(repo)

        return self.get_repository(repo.id)



from .base import Repository, PublishOptions, SyncOptions, Importer
from .container import ContainerImageRepository, ContainerSyncOptions
from .yum import YumRepository, YumSyncOptions, YumImporter
from .file import FileRepository, FileSyncOptions, FileImporter



"""Helpers for generating UD-related repo notes."""
import functools
import os
import logging
import json

from more_executors.futures import f_map, f_flat_map, f_return, f_zip

from ..model import FileUnit
from ..criteria import Criteria

LOG = logging.getLogger("pubtools.pulplib")

UD_MAPPINGS_NOTE = os.getenv("PULP_UD_MAPPINGS_NOTE") or "ud_file_release_mappings_2"


class MappingsHelper(object):
    """A wrapper for the raw release mappings dict.

    This wrapper provides a utility function for adding and removing items from the dict
    while keeping track of whether any changes have been made.
    """

    def __init__(self, data):
        self._data = data
        self.filenames_from_pulp = []  # Files for which the mapping should exist.
        self.changed = False

    def set_file_mapping(self, version, filename, order):
        """Ensure a mapping exists in the dict for a given
        version, filename & order.

        Order can be None if no specific order is required.

        Sets self.changed to True if any changes occur.
        """

        # Example of the structure we're updating:
        #
        # {
        #   "4.8.10": [
        #     {
        #     "filename": "oc-4.8.10-linux.tar.gz",
        #     "order": 1.0
        #     },
        #     {
        #     "filename": "oc-4.8.10-macosx.zip",
        #     "order": 3.0
        #     },
        #     ...,
        #   ],
        #   "4.8.11": [ ... ],
        #   ...,
        # }
        if self._data.get(version) is None:
            self._data[version] = []
            self.changed = True

        file_list = self._data[version]

        file_dict = None
        for elem in file_list:
            if elem.get("filename") == filename:
                file_dict = elem
                break
        else:
            file_dict = {"filename": filename}
            file_list.append(file_dict)
            self.changed = True

        if order is None:
            # When asked to set an order of None we'll interpret it as
            # a request to make no changes at all (rather than actually
            # setting the order to none). This keeps the possibility open
            # of ignoring pulplib logic and manually setting the order
            # values to something else by standalone tools, without pulplib
            # resetting the values every time a publish happens.
            #
            # It's intentional to make changes up to this point rather than
            # returning early if order is None, because we need to ensure
            # that filenames are put under the right 'version' even if there
            # is no 'order'.
            return

        if order == file_dict.get("order"):
            # Nothing to be changed
            return

        file_dict["order"] = order
        self.changed = True

    def remove_file_mappings(self):
        """Remove mappings for filenames that are not in Pulp."""
        for version in set(self._data):
            file_list = self._data[version]
            for i in reversed(range(len(file_list))):
                if file_list[i].get("filename") not in self.filenames_from_pulp:
                    del file_list[i]
                    self.changed = True
            # Remove empty version fields.
            if not self._data[version]:
                del self._data[version]
                self.changed = True

    @property
    def as_json(self):
        """The mapping converted to JSON form, suitable for storing on Pulp."""
        return json.dumps(self._data, sort_keys=True)


def compile_ud_mappings(repo, do_request):
    """Perform the UD mappings note compilation & update process for a given repo.

    Arguments:
        repo (~pulplib.FileRepository)
            A repository.
        do_request (callable)
            A function which can be invoked to perform an HTTP request to Pulp.

    Returns:
        A Future, resolved when the update completes successfully.
    """
    LOG.debug("%s: compiling %s", repo.id, UD_MAPPINGS_NOTE)

    # 1. Get current mappings.
    #
    # Requires a fresh retrieval of the repo since we don't store
    # these mappings on our model.
    #
    repo_url = "pulp/api/v2/repositories/%s/" % repo.id

    repo_raw_f = do_request(repo_url, method="GET")
    mappings_f = f_map(
        repo_raw_f, lambda data: (data.get("notes") or {}).get(UD_MAPPINGS_NOTE) or "{}"
    )

    # Mappings are stored as JSON, so decode them
    mappings_f = f_map(mappings_f, json.loads)

    # Wrap them in our helper for keeping track of changes
    mappings_f = f_map(mappings_f, MappingsHelper)

    # 2. Iterate over all files in the repo
    files_f = repo.search_content(Criteria.with_unit_type(FileUnit))

    # 3. Mutate the mappings as needed for each file
    updated_mappings_f = f_flat_map(
        f_zip(mappings_f, files_f),
        lambda tup: update_mappings_for_files(tup[0], tup[1]),
    )

    # 4. Upload them back if any changes
    handle_changes = functools.partial(
        upload_changed_mappings, repo=repo, repo_url=repo_url, do_request=do_request
    )
    return f_flat_map(updated_mappings_f, handle_changes)


def upload_changed_mappings(mappings, repo, repo_url, do_request):
    """Upload mappings back to repo, if and only if they've been changed."""

    if not mappings.changed:
        LOG.debug("%s: no changes required to %s", repo.id, UD_MAPPINGS_NOTE)
        return f_return()

    body = {"delta": {"notes": {UD_MAPPINGS_NOTE: mappings.as_json}}}

    # Note: Pulp docs are a bit ambiguous with respect to whether a repo PUT
    # will generate a task or not. In fact it generates tasks only if a
    # distributor or importer is being updated. Hence the request here doesn't
    # return a specific value, only a Future which succeeds if request
    # succeeds.
    update_f = do_request(repo_url, method="PUT", json=body)

    return f_map(
        update_f, lambda _: LOG.info("Updated %s in %s", UD_MAPPINGS_NOTE, repo.id)
    )


def update_mappings_for_files(mappings, file_page):
    # Add missing mappings for every file in a single page, plus all
    # following pages (async). Once all pages have been processed,
    # obsolete mappings are removed.
    #
    # Returns Future[mappings] once all pages are processed.

    for unit in file_page.data:
        # Save all the files for which the mapping should exist
        mappings.filenames_from_pulp.append(unit.path)
        version = unit.version
        if version:
            # Add missing mappings
            mappings.set_file_mapping(version, unit.path, unit.display_order)

    if not file_page.next:
        # All pages have been processed, remove mappings for files which were not listed.
        mappings.remove_file_mappings()
        # No more files, just return the mappings
        return f_return(mappings)

    # There's more files, keep going to the next page.
    return f_flat_map(
        file_page.next, lambda page: update_mappings_for_files(mappings, page)
    )
